# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for showing the testcase detail page."""

import datetime
import html
import re

from flask import request
import jinja2

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.build_management import revisions
from clusterfuzz._internal.build_management import source_mapper
from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.crash_analysis import severity_analyzer
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.fuzzing import leak_blacklist
from clusterfuzz._internal.google_cloud_utils import blobs
from clusterfuzz._internal.metrics import crash_stats
from clusterfuzz._internal.system import environment
from handlers import base_handler
from libs import access
from libs import auth
from libs import form
from libs import handler
from libs import helpers
from libs.issue_management import issue_tracker_utils

FIND_SIMILAR_ISSUES_OPTIONS = [{
    'type': 'open',
    'label': 'Open'
}, {
    'type': 'all',
    'label': 'All'
}]

CRASH_STATE_REGEXES = [
    r'bad-cast to (.+) from.*', r'bad-cast to (.+)', r'(.+) +in +.+'
]

COMPILED_CRASH_STATE_REGEXES = [
    re.compile(r, flags=re.IGNORECASE) for r in CRASH_STATE_REGEXES
]

# [<ffffff900808f5ac>] dump_backtrace+0x0/0x34c
# http://go/pakernel/msm-google/+/40e9b2ff3a280a8775cfcd5841e530ce78f94355/
# arch/arm64/kernel/traps.c#96;msm-google/arch/arm64/kernel/traps.c
KERNEL_LINK_REGEX = re.compile(
    r'(.+)(http:\/\/go\/pakernel\/[^;]+);([^;]+);(.*)')
KERNEL_LINK_FORMAT = r'%s<a href="%s">%s</a>%s'


def _parse_suspected_cls(predator_result):
  """Parse raw suspected_cls into dict."""
  if not predator_result:
    return None

  # The raw result contains some additional information that we don't need here.
  # Everything we're concerned with is a part of the "result" object included
  # with the response.
  predator_result = predator_result['result']
  return {
      'found': predator_result.get('found'),
      'suspected_project': predator_result.get('suspected_project'),
      'suspected_components': predator_result.get('suspected_components'),
      'changelists': predator_result.get('suspected_cls'),
      'feedback_url': predator_result.get('feedback_url'),
      'error_message': predator_result.get('error_message'),
  }


def highlight_common_stack_frames(crash_stacktrace):
  """Highlights common stack frames between first two stacks."""
  crash_stacks = [[]]
  highlighted_crash_stacktrace_lines = []
  old_frame_no = 0
  stack_index = 0
  stack_trace_line_format = '^ *#([0-9]+) *0x[0-9a-f]+ (.*)'

  for line in crash_stacktrace.splitlines():
    # Stacktrace separator prefix.
    if stack_index and line.startswith('+-'):
      break

    match = re.match(stack_trace_line_format, line)
    if match:
      frame_no = int(match.group(1))

      # This means we encountered another stack like free or alloc stack.
      if old_frame_no > frame_no:
        stack_index += 1
        crash_stacks.append([])

      crash_stacks[stack_index].append(match.group(2))
      old_frame_no = frame_no

  # If we have just one crash stack and no other stack,
  # then nothing to highlight.
  if stack_index == 0:
    return crash_stacktrace

  # Compare stack frames between first two stacks.
  match_index = -1
  start_index_crash_stack_1 = len(crash_stacks[0]) - 1
  start_index_crash_stack_2 = len(crash_stacks[1]) - 1
  while True:
    if (crash_stacks[0][start_index_crash_stack_1] !=
        crash_stacks[1][start_index_crash_stack_2]):
      break

    match_index = [start_index_crash_stack_1, start_index_crash_stack_2]

    if not start_index_crash_stack_1:
      break
    if not start_index_crash_stack_2:
      break

    start_index_crash_stack_1 -= 1
    start_index_crash_stack_2 -= 1

  # No match found, nothing to highlight.
  if match_index == -1:
    return crash_stacktrace

  old_frame_no = 0
  stack_index = 0
  frame_index = -1
  for line in crash_stacktrace.splitlines():
    match = re.match(stack_trace_line_format, line)
    if match:
      frame_no = int(match.group(1))

      # This means we encountered another stack like free or alloc stack.
      if old_frame_no > frame_no:
        stack_index += 1
        frame_index = -1

      frame_index += 1
      old_frame_no = frame_no

      # We only care about highlighting the first two stacks.
      if stack_index <= 1 and frame_index >= match_index[stack_index]:
        line = '<b>%s</b>' % line

    highlighted_crash_stacktrace_lines.append(line)

  return '\n'.join(highlighted_crash_stacktrace_lines)


def _linkify_android_kernel_stack_frame_if_needed(line):
  """Linkify links to android kernel source."""
  match = KERNEL_LINK_REGEX.match(line)
  if match:
    return KERNEL_LINK_FORMAT % (match.group(1), match.group(2), match.group(3),
                                 match.group(4))

  return line


def filter_stacktrace(crash_stacktrace, crash_type, revisions_dict, platform,
                      job_type):
  """Clean up and format a stack trace for display."""
  if not crash_stacktrace:
    return ''

  filtered_crash_lines = []
  for line in crash_stacktrace.splitlines():
    # Html escape line content to prevent XSS.
    line = html.escape(line, quote=True)
    line = source_mapper.linkify_stack_frame(line, revisions_dict)

    if 'android' in platform or environment.is_lkl_job(job_type):
      line = _linkify_android_kernel_stack_frame_if_needed(line)

    filtered_crash_lines.append(line)

  filtered_crash_stacktrace = '\n'.join(filtered_crash_lines)

  if crash_type == leak_blacklist.DIRECT_LEAK_LABEL:
    return leak_blacklist.highlight_first_direct_leak(filtered_crash_stacktrace)

  return highlight_common_stack_frames(filtered_crash_stacktrace)


class Line(object):
  """Represent a stacktrace line."""

  def __init__(self, line_number, content, important):
    self.line_number = line_number
    self.content = content
    self.important = important

  def __str__(self):
    return 'Line(%d, "%s", %s)' % (self.line_number, self.content,
                                   self.important)

  def __eq__(self, other):
    return hash(self) == hash(other)

  def __hash__(self):
    return hash(self.__str__())

  def to_dict(self):
    return {
        'lineNumber': self.line_number,
        'content': self.content,
        'important': self.important,
        'type': 'Line'
    }


class Gap(object):
  """Represent a gap in a previewed stacktrace."""

  def __init__(self, size):
    self.size = size

  def __str__(self):
    return 'Gap(%d)' % self.size

  def __eq__(self, other):
    return hash(self) == hash(other)

  def __hash__(self):
    return hash(self.__str__())

  def to_dict(self):
    return {'type': 'Gap', 'size': self.size}


def _is_line_important(line_content, frames):
  """Check if the line contains a frame; it means the line is
     important."""
  for frame in frames:
    if frame in line_content:
      return True
  return False


def get_stack_frames(crash_state_lines):
  """Get the stack frames from the crash state. Sometimes the crash state
     contains a type of crash, e.g. 'Bad-cast to content::RenderWidget from
     content::RenderWidgetHostViewAura'. The stack frame is
     'content::RenderWidget'."""
  frames = []
  for line in crash_state_lines:
    added = False
    for regex in COMPILED_CRASH_STATE_REGEXES:
      matches = re.match(regex, line)
      if matches:
        frames.append(matches.group(1))
        added = True
        break

    if not added:
      frames.append(line)
  return frames


def convert_to_lines(raw_stacktrace, crash_state_lines, crash_type):
  """Convert an array of string to an array of Line."""
  if not raw_stacktrace or not raw_stacktrace.strip():
    return []

  raw_lines = raw_stacktrace.splitlines()

  frames = get_stack_frames(crash_state_lines)
  escaped_frames = [jinja2.escape(f) for f in frames]
  combined_frames = frames + escaped_frames

  # Certain crash types have their own customized frames that are not related to
  # the stacktrace. Therefore, we make our best effort to preview stacktrace
  # in a reasonable way; we preview around the the top of the stacktrace.
  for unique_type in data_types.CRASH_TYPES_WITH_UNIQUE_STATE:
    if crash_type.startswith(unique_type):
      combined_frames = ['ERROR']
      break

  lines = []
  for index, content in enumerate(raw_lines):
    important = _is_line_important(content, combined_frames)
    lines.append(Line(index + 1, content, important))
  return lines


def get_testcase_detail_by_id(testcase_id):
  """Get testcase detail for rendering the testcase detail page."""
  testcase = access.check_access_and_get_testcase(testcase_id)
  return get_testcase_detail(testcase)


def _get_revision_range_html_from_string(job_type, revision_range):
  """Return revision range html for a revision range and job type given a range
  string."""
  try:
    start_revision, end_revision = revision_range.split(':')
  except:
    return 'Bad revision range.'

  return _get_revision_range_html(job_type, start_revision, end_revision)


def _get_revision_range_html(job_type, start_revision, end_revision=None):
  """Return revision range html for a revision range and job type."""
  if end_revision is None:
    end_revision = start_revision

  component_rev_list = revisions.get_component_range_list(
      start_revision, end_revision, job_type)
  if not component_rev_list:
    return ('%s:%s (No component revisions found!)' % (start_revision,
                                                       end_revision))

  return revisions.format_revision_list(component_rev_list)


def _get_blob_size_string(blob_key):
  """Return blob size string."""
  blob_size = blobs.get_blob_size(blob_key)
  if blob_size is None:
    return None

  return utils.get_size_string(blob_size)


def _format_reproduction_help(reproduction_help):
  """Format a reproduction help string as HTML (linkified with break tags)."""
  if not reproduction_help:
    return ''

  return jinja2.utils.urlize(reproduction_help).replace('\n', '<br>')


def get_testcase_detail(testcase):
  """Get testcase detail for rendering the testcase detail page."""
  config = db_config.get()
  crash_address = testcase.crash_address
  crash_state = testcase.crash_state
  crash_state_lines = crash_state.strip().splitlines()
  crash_type = data_handler.get_crash_type_string(testcase)
  external_user = not access.has_access(job_type=testcase.job_type)
  issue_url = issue_tracker_utils.get_issue_url(testcase)
  metadata = testcase.get_metadata()
  original_testcase_size = _get_blob_size_string(testcase.fuzzed_keys)
  minimized_testcase_size = _get_blob_size_string(testcase.minimized_keys)
  has_issue_tracker = bool(data_handler.get_issue_tracker_name())

  fuzzer_display = data_handler.get_fuzzer_display(testcase)

  formatted_reproduction_help = _format_reproduction_help(
      data_handler.get_formatted_reproduction_help(testcase))
  # When we have a HELP_TEMPLATE, ignore any default values set for HELP_URL.
  if not formatted_reproduction_help:
    reproduction_help_url = data_handler.get_reproduction_help_url(
        testcase, config)
  else:
    reproduction_help_url = None

  if not testcase.regression:
    regression = 'Pending'
  elif testcase.regression == 'NA':
    regression = 'NA'
  else:
    regression = _get_revision_range_html_from_string(testcase.job_type,
                                                      testcase.regression)

  fixed_full = None
  if 'progression_pending' in metadata:
    fixed = 'Pending'
  elif not testcase.fixed:
    fixed = 'NO'
  elif testcase.fixed == 'NA':
    fixed = 'NA'
  elif testcase.fixed == 'Yes':
    fixed = 'YES'
  else:
    fixed = 'YES'
    fixed_full = _get_revision_range_html_from_string(testcase.job_type,
                                                      testcase.fixed)

  last_tested = None
  last_tested_revision = (
      metadata.get('last_tested_revision') or testcase.crash_revision)
  if last_tested_revision:
    last_tested = _get_revision_range_html(testcase.job_type,
                                           last_tested_revision)

  crash_revision = testcase.crash_revision
  crash_revisions_dict = revisions.get_component_revisions_dict(
      crash_revision, testcase.job_type)
  crash_stacktrace = data_handler.get_stacktrace(testcase)
  crash_stacktrace = filter_stacktrace(crash_stacktrace, testcase.crash_type,
                                       crash_revisions_dict, testcase.platform,
                                       testcase.job_type)
  crash_stacktrace = convert_to_lines(crash_stacktrace, crash_state_lines,
                                      crash_type)

  last_tested_crash_revision = metadata.get('last_tested_crash_revision')
  last_tested_crash_revisions_dict = revisions.get_component_revisions_dict(
      last_tested_crash_revision, testcase.job_type)
  last_tested_crash_stacktrace = data_handler.get_stacktrace(
      testcase, stack_attribute='last_tested_crash_stacktrace')
  last_tested_crash_stacktrace = filter_stacktrace(
      last_tested_crash_stacktrace, testcase.crash_type,
      last_tested_crash_revisions_dict, testcase.platform, testcase.job_type)
  last_tested_crash_stacktrace = convert_to_lines(last_tested_crash_stacktrace,
                                                  crash_state_lines, crash_type)

  privileged_user = access.has_access(need_privileged_access=True)

  # Fix build url link. |storage.cloud.google.com| takes care of using the
  # right set of authentication credentials needed to access the link.
  if 'build_url' in metadata:
    metadata['build_url'] = metadata['build_url'].replace(
        'gs://', 'https://storage.cloud.google.com/')

  pending_blame_task = (
      testcase.has_blame() and 'blame_pending' in metadata and
      metadata['blame_pending'])
  pending_impact_task = (
      testcase.has_impacts() and not testcase.is_impact_set_flag)
  pending_minimize_task = not testcase.minimized_keys
  pending_progression_task = ('progression_pending' in metadata and
                              metadata['progression_pending'])
  pending_regression_task = not testcase.regression
  pending_stack_task = testcase.last_tested_crash_stacktrace == 'Pending'
  needs_refresh = (
      testcase.status == 'Pending' or
      ((testcase.status == 'Processed' or testcase.status == 'Duplicate') and
       (pending_blame_task or pending_impact_task or pending_minimize_task or
        pending_progression_task or pending_regression_task or
        pending_stack_task)))

  if data_types.SecuritySeverity.is_valid(testcase.security_severity):
    security_severity = severity_analyzer.severity_to_string(
        testcase.security_severity)
  else:
    security_severity = None

  auto_delete_timestamp = None
  auto_close_timestamp = None

  if testcase.one_time_crasher_flag:
    last_crash_time = (
        crash_stats.get_last_crash_time(testcase) or testcase.timestamp)

    # Set auto-delete timestamp for unreproducible testcases with
    # no associated bug.
    if not testcase.bug_information:
      auto_delete_timestamp = utils.utc_datetime_to_timestamp(
          last_crash_time + datetime.timedelta(
              days=data_types.UNREPRODUCIBLE_TESTCASE_NO_BUG_DEADLINE))

    # Set auto-close timestamp for unreproducible testcases with
    # an associated bug.
    if testcase.open and testcase.bug_information:
      auto_close_timestamp = utils.utc_datetime_to_timestamp(
          last_crash_time + datetime.timedelta(
              days=data_types.UNREPRODUCIBLE_TESTCASE_WITH_BUG_DEADLINE))

  memory_tool_display_string = environment.get_memory_tool_display_string(
      testcase.job_type)
  memory_tool_display_label = memory_tool_display_string.split(':')[0]
  memory_tool_display_value = memory_tool_display_string.split(':')[1].strip()

  helpers.log('Testcase %s' % testcase.key.id(), helpers.VIEW_OPERATION)
  return {
      'id':
          testcase.key.id(),
      'crash_type':
          crash_type,
      'crash_address':
          crash_address,
      'crash_state':
          crash_state,  # Used by reproduce tool.
      'crash_state_lines':
          crash_state_lines,
      'crash_revision':
          testcase.crash_revision,
      'csrf_token':
          form.generate_csrf_token(),
      'external_user':
          external_user,
      'footer':
          testcase.comments,
      'formatted_reproduction_help':
          formatted_reproduction_help,
      'fixed':
          fixed,
      'fixed_full':
          fixed_full,
      'issue_url':
          issue_url,
      'is_admin':
          auth.is_current_user_admin(),
      'metadata':
          metadata,
      'minimized_testcase_size':
          minimized_testcase_size,
      'needs_refresh':
          needs_refresh,
      'original_testcase_size':
          original_testcase_size,
      'privileged_user':
          privileged_user,
      'regression':
          regression,
      'crash_stacktrace': {
          'lines':
              crash_stacktrace,
          'revision':
              revisions.get_real_revision(
                  crash_revision, testcase.job_type, display=True)
      },
      'last_tested_crash_stacktrace': {
          'lines':
              last_tested_crash_stacktrace,
          'revision':
              revisions.get_real_revision(
                  last_tested_crash_revision, testcase.job_type, display=True)
      },
      'security_severity':
          security_severity,
      'security_severities':
          data_types.SecuritySeverity.list(),
      'stats': {
          'min_hour': crash_stats.get_min_hour(),
          'max_hour': crash_stats.get_max_hour(),
      },
      'suspected_cls':
          _parse_suspected_cls(metadata.get('predator_result')),
      'testcase':
          testcase,
      'timestamp':
          utils.utc_datetime_to_timestamp(testcase.timestamp),
      'show_blame':
          testcase.has_blame(),
      'show_impact':
          testcase.has_impacts(),
      'impacts_production':
          testcase.impacts_production(),
      'find_similar_issues_options':
          FIND_SIMILAR_ISSUES_OPTIONS,
      'auto_delete_timestamp':
          auto_delete_timestamp,
      'auto_close_timestamp':
          auto_close_timestamp,
      'memory_tool_display_label':
          memory_tool_display_label,
      'memory_tool_display_value':
          memory_tool_display_value,
      'last_tested':
          last_tested,
      'is_admin_or_not_oss_fuzz':
          is_admin_or_not_oss_fuzz(),
      'has_issue_tracker':
          has_issue_tracker,
      'reproduction_help_url':
          reproduction_help_url,
      'is_local_development':
          environment.is_running_on_app_engine_development(),
      'fuzzer_display':
          fuzzer_display._asdict(),
  }


def is_admin_or_not_oss_fuzz():
  """Return True if the current user is an admin or if this is not OSS-Fuzz."""
  return not utils.is_oss_fuzz() or auth.is_current_user_admin()


class Handler(base_handler.Handler):
  """Handler that shows a testcase in detail."""

  @handler.get(handler.HTML)
  def get(self, testcase_id=None):
    """Serve the testcase detail HTML page."""
    values = {'info': get_testcase_detail_by_id(testcase_id)}
    return self.render('testcase-detail.html', values)


class DeprecatedHandler(base_handler.Handler):
  """Deprecated handler to show old style testcase link with key."""

  def get(self):
    """Serve the redirect to the current test case detail page."""
    testcase_id = request.args.get('key')
    if not testcase_id:
      raise helpers.EarlyExitException('No testcase key provided.', 400)

    return self.redirect('/testcase-detail/%s' % testcase_id)


class RefreshHandler(base_handler.Handler):
  """Handler that shows a testcase in detail through JSON."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.oauth
  def post(self):
    """Serve the testcase detail JSON."""
    testcase_id = request.get('testcaseId')
    return self.render_json(get_testcase_detail_by_id(testcase_id))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cron to managed VMs."""

from collections import namedtuple
from concurrent.futures import ThreadPoolExecutor
import copy
import itertools
import json
import logging

from google.cloud import ndb

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.datastore import ndb_utils
from clusterfuzz._internal.google_cloud_utils import compute_engine_projects
from handlers import base_handler
from handlers.cron.helpers import bot_manager
from libs import handler

PROJECT_MIN_CPUS = 1

# This is the maximum number of instances supported in a single instance group.
PROJECT_MAX_CPUS = 1000

NUM_THREADS = 8

WorkerInstance = namedtuple('WorkerInstance', ['name', 'project'])


class ManageVmsException(Exception):
  """Base exception class."""


def _get_project_ids():
  """Return the GCE project IDs."""
  return list(local_config.Config(local_config.GCE_CLUSTERS_PATH).get().keys())


def _instance_name_from_url(instance_url):
  """Extract instance name from url."""
  return instance_url.split('/')[-1]


def get_resource_name(prefix, project_name):
  """Get a name that can be used for GCE resources."""
  # https://cloud.google.com/compute/docs/reference/latest/instanceGroupManagers
  max_name_length = 58

  project_name = project_name.lower().replace('_', '-')
  name = prefix + '-' + project_name
  return name[:max_name_length]


def get_template_body(gce_project,
                      template_name,
                      task_tag=None,
                      disk_size_gb=None,
                      service_account=None,
                      tls_cert=None):
  """Return the instance template body."""
  template_body = copy.deepcopy(
      gce_project.get_instance_template(template_name))
  if task_tag:
    template_body['properties']['metadata']['items'].append({
        'key': 'task-tag',
        'value': task_tag,
    })

  if disk_size_gb:
    disk = template_body['properties']['disks'][0]
    disk['initializeParams']['diskSizeGb'] = disk_size_gb

  if service_account:
    template_body['properties']['serviceAccounts'][0]['email'] = service_account

  if tls_cert:
    template_body['properties']['metadata']['items'].extend([{
        'key': 'tls-cert',
        'value': tls_cert.cert_contents.decode('utf-8'),
    }, {
        'key': 'tls-key',
        'value': tls_cert.key_contents.decode('utf-8'),
    }])

  return template_body


def _get_template_disk_size(template):
  """Get disk size from template."""
  return int(
      template['properties']['disks'][0]['initializeParams']['diskSizeGb'])


def _get_template_service_account(template):
  """Get service account from template."""
  return template['properties']['serviceAccounts'][0]['email']


def _get_metadata_value(metadata_items, key):
  return next((item['value'] for item in metadata_items if item['key'] == key),
              None)


def _template_needs_update(current_template, new_template, resource_name):
  """Return whether or not the template needs an update."""
  current_version = json.loads(current_template['description'])['version']
  new_version = json.loads(new_template['description'])['version']

  if current_version != new_version:
    logging.info(
        'Instance template version out of date '
        '(current=%s, new=%s): %s', current_version, new_version, resource_name)
    return True

  current_disk_size_gb = _get_template_disk_size(current_template)
  new_disk_size_gb = _get_template_disk_size(new_template)

  if current_disk_size_gb != new_disk_size_gb:
    logging.info(
        'Instance template disk size changed '
        '(current=%d, new=%d): %s', current_disk_size_gb, new_disk_size_gb,
        resource_name)
    return True

  current_service_account = _get_template_service_account(current_template)
  new_service_account = _get_template_service_account(new_template)

  if current_service_account != new_service_account:
    logging.info('Service account changed '
                 '(current=%s, new=%s): %s', current_service_account,
                 new_service_account, resource_name)
    return True

  current_tls_cert = _get_metadata_value(
      current_template['properties']['metadata']['items'], 'tls-cert')

  new_tls_cert = _get_metadata_value(
      new_template['properties']['metadata']['items'], 'tls-cert')

  if current_tls_cert != new_tls_cert:
    logging.info('TLS cert changed.')
    return True

  return False


class ClustersManager(object):
  """Manager for clusters in a project."""

  def __init__(self, project_id):
    self.gce_project = compute_engine_projects.load_project(project_id)
    self.thread_pool = None
    self.pending_updates = []

  def start_thread_pool(self):
    """Start the thread pool."""
    self.thread_pool = ThreadPoolExecutor(max_workers=NUM_THREADS)

  def wait_updates(self):
    """Wait for updates to finish."""
    for update in self.pending_updates:
      # Raise any exceptions.
      update.result()

    self.pending_updates = []

  def finish_updates(self):
    """Close the thread pool and finish cluster updates."""
    self.wait_updates()
    self.thread_pool.shutdown()
    self.thread_pool = None

  def update_clusters(self):
    """Update all clusters in a project."""
    self.start_thread_pool()

    for cluster in self.gce_project.clusters:
      self.pending_updates.append(
          self.thread_pool.submit(self.update_cluster, cluster, cluster.name,
                                  cluster.instance_count))

    self.finish_updates()

  def update_cluster(self,
                     cluster,
                     resource_name,
                     cpu_count,
                     task_tag=None,
                     disk_size_gb=None,
                     service_account=None,
                     tls_cert=None):
    """Update the cluster."""
    manager = bot_manager.BotManager(self.gce_project.project_id,
                                     cluster.gce_zone)

    instance_template = manager.instance_template(resource_name)
    instance_group = manager.instance_group(resource_name)

    # Load expected template body.
    template_body = get_template_body(
        self.gce_project,
        cluster.instance_template,
        task_tag=task_tag,
        disk_size_gb=disk_size_gb,
        service_account=service_account,
        tls_cert=tls_cert)

    if instance_template.exists():
      # Check for updates.
      current_template_body = instance_template.get()
      template_needs_update = _template_needs_update(
          current_template_body, template_body, resource_name)
    else:
      logging.info('Creating new instance template: %s', resource_name)
      instance_template.create(template_body)
      template_needs_update = False

    if instance_group.exists():
      if template_needs_update:
        # Instance groups need to be deleted first before an instance template
        # can be deleted.
        logging.info('Deleting instance group %s for template update.',
                     resource_name)
        try:
          instance_group.delete()
        except bot_manager.NotFoundError:
          # Already deleted.
          pass
      else:
        instance_group_body = instance_group.get()
        if instance_group_body['targetSize'] != cpu_count:
          logging.info('Resizing instance group %s from %d to %d.',
                       resource_name, instance_group_body['targetSize'],
                       cpu_count)
          try:
            instance_group.resize(cpu_count, wait_for_instances=False)
          except bot_manager.OperationError as e:
            logging.error('Failed to resize instance group %s: %s',
                          resource_name, str(e))

        else:
          logging.info('No instance group size changes needed.')

        return

    if template_needs_update:
      logging.info('Recreating instance template: %s', resource_name)
      instance_template.delete()
      instance_template.create(template_body)

    logging.info('Creating new instance group: %s', resource_name)
    try:
      instance_group.create(
          resource_name,
          resource_name,
          size=cpu_count,
          wait_for_instances=False)
    except bot_manager.OperationError as e:
      logging.error('Failed to create instance group %s: %s', resource_name,
                    str(e))


class OssFuzzClustersManager(ClustersManager):
  """Manager for clusters in OSS-Fuzz."""

  def __init__(self, project_id):
    super().__init__(project_id)
    self.worker_to_assignment = {}
    for assignment in self.gce_project.host_worker_assignments:
      self.worker_to_assignment[assignment.worker] = assignment

    self.all_host_names = set()

  def update_clusters(self):
    """Update all clusters in a project."""
    self.start_thread_pool()

    all_projects = list(data_types.OssFuzzProject.query().order(
        data_types.OssFuzzProject.name))

    self.cleanup_old_projects([project.name for project in all_projects])

    projects = [project for project in all_projects if not project.high_end]
    high_end_projects = [
        project for project in all_projects if project.high_end
    ]

    project_infos = [
        self.get_or_create_project_info(project.name) for project in projects
    ]

    high_end_project_infos = [
        self.get_or_create_project_info(project.name)
        for project in high_end_projects
    ]

    for project, project_info in itertools.chain(
        list(zip(projects, project_infos)),
        list(zip(high_end_projects, high_end_project_infos))):
      self.cleanup_clusters(project, project_info)

    for cluster in self.gce_project.clusters:
      self.update_project_cpus(projects, project_infos, high_end_projects,
                               high_end_project_infos, cluster)

    self.cleanup_old_assignments(self.all_host_names)
    self.finish_updates()

  def get_or_create_project_info(self, project_name):
    """Get OSS-Fuzz CPU info by project name (or create a new one if it doesn't
    exist)."""
    key = ndb.Key(data_types.OssFuzzProjectInfo, project_name)
    project_info = key.get()
    if not project_info:
      project_info = data_types.OssFuzzProjectInfo(
          name=project_name, id=project_name)
      project_info.put()

    return project_info

  def get_or_create_host_worker_assignment(self, host_name, instance_num):
    """Get OSS-Fuzz host worker assignment (or create a new one if it doesn't
    exist)."""
    key_id = '%s-%d' % (host_name, instance_num)
    key = ndb.Key(data_types.HostWorkerAssignment, key_id)
    assignment = key.get()
    if not assignment:
      assignment = data_types.HostWorkerAssignment(
          host_name=host_name, instance_num=instance_num, id=key_id)
      assignment.put()

    return assignment

  def cleanup_old_assignments(self, host_names):
    """Remove old OSS-Fuzz host worker assignment entries."""
    to_delete = []
    for assignment in data_types.HostWorkerAssignment.query():
      if assignment.host_name not in host_names:
        to_delete.append(assignment.key)

    ndb_utils.delete_multi(to_delete)

  def distribute_cpus(self, projects, total_cpus):
    """Distribute OSS-Fuzz CPUs for each project by weight.

    |projects| should be sorted
    alphabetically by name to ensure determinism for the same set of CPUs.
    """
    available_cpus = total_cpus
    total_weight = sum(project.cpu_weight for project in projects)

    cpu_count = []

    for project in projects:
      if total_weight:
        share = project.cpu_weight / total_weight
      else:
        share = 0.0

      share_cpus = int(total_cpus * share)
      share_cpus = max(PROJECT_MIN_CPUS, share_cpus)
      share_cpus = min(PROJECT_MAX_CPUS, share_cpus)

      if share_cpus <= available_cpus:
        cpu_count.append(share_cpus)
        available_cpus -= share_cpus
      else:
        cpu_count.append(0)

    # indexes into |project| sorted by highest weight first.
    indexes_by_weight = sorted(
        list(range(len(projects))),
        key=lambda k: projects[k].cpu_weight,
        reverse=True)

    # Distribute the remainder from rounding errors (and capping) up to the cap,
    # preferring projects with a higher weight first.
    while available_cpus:
      cpus_allocated = 0

      for i in range(len(cpu_count)):
        project_index = indexes_by_weight[i]

        if cpu_count[project_index] < PROJECT_MAX_CPUS:
          cpu_count[project_index] += 1
          cpus_allocated += 1

        if cpus_allocated >= available_cpus:
          break

      if not cpus_allocated:
        # Hit the cap for each project. Realistically, this shouldn't ever
        # happen.
        break

      available_cpus -= cpus_allocated

    if available_cpus:
      logging.warning('%d CPUs are not being used.', available_cpus)

    return cpu_count

  def do_assign_hosts_to_workers(self, host_names, worker_instances,
                                 workers_per_host):
    """Assign OSS-Fuzz host instances to workers."""
    # Sort host and worker instance names to make assignment deterministic for
    # the same initial set of host and workers.
    host_names.sort()
    worker_instances.sort(key=lambda w: w.name)

    # Algorithm:
    # For each host instance,
    #   - If there is already an assignment, and a worker with the same name
    #   still exists, do nothing.
    #   - Otherwise, assign it to the first unassigned worker (in alphabetical
    #   order).
    # This should ensure that a worker is reassigned only if it was
    # reimaged/new.
    current_worker_names = {worker.name for worker in worker_instances}
    previous_assigned_workers = set()

    new_assignments = []

    for host_name in host_names:
      for i in range(0, workers_per_host):
        assignment = self.get_or_create_host_worker_assignment(host_name, i)
        if (assignment.worker_name and
            assignment.worker_name in current_worker_names):
          # Existing assignment is still valid. Don't do anything for these.
          logging.info('Keeping old assignment of %s(%d) -> %s.', host_name, i,
                       assignment.worker_name)
          previous_assigned_workers.add(assignment.worker_name)
          continue

        # This host instance was either unassigned or the worker it was
        # connected to no longer exists, so we need to assign it to a new
        # worker.
        new_assignments.append(assignment)

    new_workers = [
        worker for worker in worker_instances
        if worker.name not in previous_assigned_workers
    ]

    assert len(new_assignments) == len(new_workers)
    for assignment, worker in zip(new_assignments, new_workers):
      assignment.worker_name = worker.name
      assignment.project_name = worker.project
      logging.info('New assignment: %s(%d) - >%s.', assignment.host_name,
                   assignment.instance_num, assignment.worker_name)

    return new_assignments

  def delete_gce_resources(self, project_info, cluster_info):
    """Delete instance templates and instance groups."""
    manager = bot_manager.BotManager(self.gce_project.project_id,
                                     cluster_info.gce_zone)

    resource_name = get_resource_name(cluster_info.cluster, project_info.name)

    try:
      manager.instance_group(resource_name).delete()
    except bot_manager.NotFoundError:
      logging.info('Instance group %s already deleted.', resource_name)

    try:
      manager.instance_template(resource_name).delete()
    except bot_manager.NotFoundError:
      logging.info('Instance template %s already deleted.', resource_name)

  def cleanup_old_projects(self, existing_project_names):
    """Cleanup old projects."""
    to_delete = []

    for project_info in list(data_types.OssFuzzProjectInfo.query()):
      if project_info.name in existing_project_names:
        continue

      logging.info('Deleting %s', project_info.name)

      for cluster_info in project_info.clusters:
        self.delete_gce_resources(project_info, cluster_info)

      to_delete.append(project_info.key)

    ndb_utils.delete_multi(to_delete)

  def cleanup_clusters(self, project, project_info):
    """Remove nonexistant clusters."""
    existing_cluster_names = [
        cluster.name for cluster in self.gce_project.clusters
    ]

    # Delete clusters that no longer exist, or the if the high end flag changed
    # for a project.
    to_delete = [
        cluster_info for cluster_info in project_info.clusters if
        (cluster_info.cluster not in existing_cluster_names or project.high_end
         != self.gce_project.get_cluster(cluster_info.cluster).high_end)
    ]
    if not to_delete:
      return

    for cluster_info in to_delete:
      logging.info('Deleting old cluster %s for %s.', cluster_info.cluster,
                   project_info.name)
      self.delete_gce_resources(project_info, cluster_info)

    project_info.clusters = [
        cluster_info for cluster_info in project_info.clusters
        if cluster_info.cluster in existing_cluster_names
    ]

  def update_project_cluster(self,
                             project,
                             project_info,
                             cluster,
                             cpu_count,
                             disk_size_gb=None):
    """Update cluster allocation for a project."""
    service_account = None
    tls_cert = None

    if cluster.worker:
      # If this cluster is for untrusted workers, use the project service
      # account.
      service_account = project.service_account
      tls_cert = ndb.Key(data_types.WorkerTlsCert, project.name).get()
      if not tls_cert:
        logging.warning('TLS certs not set up yet for %s.', project.name)
        return

    cluster_info = project_info.get_cluster_info(cluster.name)
    if not cluster_info:
      project_info.clusters.append(
          data_types.OssFuzzProjectInfo.ClusterInfo(
              cluster=cluster.name,
              gce_zone=cluster.gce_zone,
              cpu_count=cpu_count))
      cluster_info = project_info.clusters[-1]

    # Get a name that can be used for the instance template and instance group.
    resource_name = get_resource_name(cluster.name, project_info.name)

    def do_update():
      """Update the cluster and cpu count info."""
      self.update_cluster(
          cluster,
          resource_name,
          cpu_count,
          task_tag=project_info.name,
          disk_size_gb=disk_size_gb,
          service_account=service_account,
          tls_cert=tls_cert)

      cluster_info.cpu_count = cpu_count

    self.pending_updates.append(self.thread_pool.submit(do_update))

  def update_project_cpus(self, projects, project_infos, high_end_projects,
                          high_end_project_infos, cluster):
    """Update CPU allocations for each project."""
    # Calculate CPUs in each cluster.
    if not cluster.distribute:
      self.pending_updates.append(
          self.thread_pool.submit(self.update_cluster, cluster, cluster.name,
                                  cluster.instance_count))
      return

    if cluster.high_end:
      current_projects = high_end_projects
      current_project_infos = high_end_project_infos
    else:
      current_projects = projects
      current_project_infos = project_infos

    cpu_counts = self.distribute_cpus(current_projects, cluster.instance_count)

    # Resize projects starting with ones that reduce number of CPUs. This is
    # so that we always have quota when we're resizing a project cluster.
    # pylint: disable=cell-var-from-loop
    def _cpu_diff_key(index):
      cluster_info = current_project_infos[index].get_cluster_info(cluster.name)
      if cluster_info and cluster_info.cpu_count is not None:
        old_cpu_count = cluster_info.cpu_count
      else:
        old_cpu_count = 0

      return cpu_counts[index] - old_cpu_count

    resize_order = sorted(list(range(len(cpu_counts))), key=_cpu_diff_key)
    for i in resize_order:
      project = current_projects[i]
      project_info = current_project_infos[i]
      self.update_project_cluster(
          project,
          project_info,
          cluster,
          cpu_counts[i],
          disk_size_gb=project.disk_size_gb)

    self.wait_updates()
    ndb_utils.put_multi(project_infos)
    ndb_utils.put_multi(high_end_project_infos)

    # If the workers are done, we're ready to assign them.
    # Note: This assumes that hosts are always specified before workers.
    if cluster.name in self.worker_to_assignment:
      self.assign_hosts_to_workers(self.worker_to_assignment[cluster.name])

  def get_all_workers_in_cluster(self, manager, cluster_name):
    """Get all workers in a cluster."""
    workers = []
    project_infos = list(data_types.OssFuzzProjectInfo.query().order(
        data_types.OssFuzzProjectInfo.name))

    for project_info in project_infos:
      cluster_info = next((cluster for cluster in project_info.clusters
                           if cluster.cluster == cluster_name), None)
      if not cluster_info or cluster_info.cpu_count == 0:
        continue

      worker_group_name = get_resource_name(cluster_info.cluster,
                                            project_info.name)
      worker_instance_group = manager.instance_group(worker_group_name)
      if not worker_instance_group.exists():
        logging.error('Worker instance group %s does not exist.',
                      worker_group_name)
        continue

      instances = list(worker_instance_group.list_managed_instances())
      if len(instances) != cluster_info.cpu_count:
        logging.error(
            'Number of instances in instance group %s does not match.'
            'Expected %d, got %d.', worker_group_name, cluster_info.cpu_count,
            len(instances))
        raise ManageVmsException('Inconsistent instance count in group.')

      for instance in instances:
        workers.append(
            WorkerInstance(
                name=_instance_name_from_url(instance['instance']),
                project=project_info.name))

    return workers

  def assign_hosts_to_workers(self, assignment):
    """Assign host instances to workers."""
    host_cluster = self.gce_project.get_cluster(assignment.host)
    worker_cluster = self.gce_project.get_cluster(assignment.worker)

    if host_cluster.gce_zone != worker_cluster.gce_zone:
      logging.error('Mismatching zones for %s and %s.', assignment.host,
                    assignment.worker)
      return

    if (host_cluster.instance_count * assignment.workers_per_host !=
        worker_cluster.instance_count):
      logging.error('Invalid host/worker cluster size for %s and %s.',
                    assignment.host, assignment.worker)
      return

    if host_cluster.high_end != worker_cluster.high_end:
      logging.error('Mismatching high end setting for %s and %s',
                    assignment.host, assignment.worker)
      return

    manager = bot_manager.BotManager(self.gce_project.project_id,
                                     host_cluster.gce_zone)
    host_instance_group = manager.instance_group(host_cluster.name)

    if not host_instance_group.exists():
      logging.error('Host instance group %s does not exist.', host_cluster.name)
      return

    host_names = [
        _instance_name_from_url(instance['instance'])
        for instance in host_instance_group.list_managed_instances()
    ]
    self.all_host_names.update(host_names)
    worker_instances = self.get_all_workers_in_cluster(manager,
                                                       worker_cluster.name)

    if len(worker_instances) != worker_cluster.instance_count:
      logging.error(
          'Actual number of worker instances for %s did not match. '
          'Expected %d, got %d.', worker_cluster.name,
          worker_cluster.instance_count, len(worker_instances))
      return

    new_assignments = self.do_assign_hosts_to_workers(
        host_names, worker_instances, assignment.workers_per_host)
    ndb_utils.put_multi(new_assignments)


class Handler(base_handler.Handler):
  """CPU distributor for OSS-Fuzz projects."""

  @handler.cron()
  def get(self):
    """Handle a get request."""
    if utils.is_oss_fuzz():
      manager_class = OssFuzzClustersManager
    else:
      manager_class = ClustersManager

    for project_id in _get_project_ids():
      manager = manager_class(project_id)
      manager.update_clusters()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for bot_manager."""

import getpass
import unittest

from handlers.cron.helpers import bot_manager

TEST_PROJECT = 'clusterfuzz-testing'
TEST_ZONE = 'us-central1-b'
TEST_INSTANCE_TEMPLATE = 'instance-template-'
TEST_INSTANCE_GROUP = 'instance-group-'


def test_template():
  return {
      'name': test_instance_template_name(),
      'description': 'test',
      'properties': {
          'machineType':
              'n1-standard-1',
          'disks': [{
              'boot': True,
              'initializeParams': {
                  'sourceImage': ('projects/cos-cloud/global/images/family/'
                                  'cos-stable'),
                  'diskSizeGb':
                      10,
                  'diskType':
                      'pd-standard'
              }
          }],
          'networkInterfaces': [{
              'network': 'global/networks/default',
          }]
      }
  }


def test_instance_group_name():
  return TEST_INSTANCE_GROUP + getpass.getuser()


def test_instance_template_name():
  return TEST_INSTANCE_TEMPLATE + getpass.getuser()


def cleanup_resources():
  """Clean up resources."""
  manager = bot_manager.BotManager(TEST_PROJECT, TEST_ZONE)
  try:
    manager.instance_group(test_instance_group_name()).delete()
  except bot_manager.NotFoundError:
    pass

  try:
    manager.instance_template(test_instance_template_name()).delete()
  except bot_manager.NotFoundError:
    pass


@unittest.skip('Slow integration test')
class BotManagerTest(unittest.TestCase):
  """Tests for bot_manager."""

  def setUp(self):
    cleanup_resources()
    self.manager = bot_manager.BotManager(TEST_PROJECT, TEST_ZONE)

  def tearDown(self):
    cleanup_resources()

  def test_templates(self):
    """Test instance templates."""
    template = self.manager.instance_template(test_instance_template_name())
    self.assertFalse(template.exists())
    with self.assertRaises(bot_manager.NotFoundError):
      template.get()

    expected_properties = {
        'disks': [{
            'boot': True,
            'initializeParams': {
                'diskSizeGb':
                    '10',
                'diskType':
                    'pd-standard',
                'sourceImage': ('projects/cos-cloud/global/images/family/'
                                'cos-stable')
            },
            'kind': 'compute#attachedDisk',
            'mode': 'READ_WRITE',
            'type': 'PERSISTENT'
        }],
        'machineType':
            'n1-standard-1',
        'networkInterfaces': [{
            'kind':
                'compute#networkInterface',
            'network': ('https://www.googleapis.com/compute/v1/projects/'
                        '%s/global/networks/default' % TEST_PROJECT),
        }],
    }

    template.create(test_template())
    self.assertTrue(template.exists())

    body = template.get()
    self.assertEqual(body['name'], test_instance_template_name())
    self.assertEqual(body['description'], 'test')
    self.assertDictEqual(expected_properties, body['properties'])

    template.delete()
    self.assertFalse(template.exists())

  def test_instance_groups(self):
    """Test instance groups."""
    template = self.manager.instance_template(test_instance_template_name())
    template.create(test_template())

    group = self.manager.instance_group(test_instance_group_name())

    self.assertFalse(group.exists())
    with self.assertRaises(bot_manager.NotFoundError):
      group.get()

    group.create(test_instance_group_name(), template.name, 1)
    self.assertTrue(group.exists())

    body = group.get()
    self.assertEqual(
        'https://www.googleapis.com/compute/v1/projects/'
        '%s/global/instanceTemplates/%s' % (TEST_PROJECT,
                                            test_instance_template_name()),
        body['instanceTemplate'])
    self.assertEqual(1, body['targetSize'])

    managed_instances = list(group.list_managed_instances())
    self.assertEqual(1, len(managed_instances))
    self.assertEqual('RUNNING', managed_instances[0]['instanceStatus'])
    self.assertEqual('NONE', managed_instances[0]['currentAction'])
    self.assertTrue(managed_instances[0]['instance'].startswith(
        'https://www.googleapis.com/compute/v1/projects/'
        '%s/zones/%s/instances/%s' % (TEST_PROJECT, TEST_ZONE,
                                      test_instance_group_name())))

    group.resize(2)
    body = group.get()
    self.assertEqual(2, body['targetSize'])
    managed_instances = list(group.list_managed_instances())
    self.assertEqual(2, len(managed_instances))

    group.delete()
    self.assertFalse(group.exists())



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""show tests."""
# pylint: disable=protected-access
import collections
import datetime
import os
import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.testcase_detail import show
from libs import helpers


class ParseSuspectedClsTest(unittest.TestCase):
  """Test _parse_suspected_cls."""

  def test_empty(self):
    """Test when we have no predator result."""
    self.assertIsNone(show._parse_suspected_cls(None))

  def test_all_present(self):
    """Ensure that we work properly when all expected fields are present."""
    raw_predator_result = {}
    raw_predator_result['result'] = {
        'found': True,
        'suspected_project': 'suspected-project',
        'suspected_components': ['A', 'B>C'],
        'suspected_cls': ['dummy cl'],
        'feedback_url': 'https://feedback/',
        'error_message': 'error string',
        'unused': 'unused',
    }

    # Note that this doesn't contain the "unused" field.
    expected_parsed_result = {
        'found': True,
        'suspected_project': 'suspected-project',
        'suspected_components': ['A', 'B>C'],
        'changelists': ['dummy cl'],
        'feedback_url': 'https://feedback/',
        'error_message': 'error string',
    }

    self.assertDictEqual(
        show._parse_suspected_cls(raw_predator_result), expected_parsed_result)

  def test_all_missing(self):
    """Ensure that we work properly if no expected fields are present."""
    raw_predator_result = {'result': {}}
    expected_parsed_result = {
        'found': None,
        'suspected_project': None,
        'suspected_components': None,
        'changelists': None,
        'feedback_url': None,
        'error_message': None,
    }

    self.assertDictEqual(
        show._parse_suspected_cls(raw_predator_result), expected_parsed_result)


class GetStackFramesTest(unittest.TestCase):
  """Test get_stack_frames."""

  def test_normal(self):
    """Test getting normal crash state."""
    self.assertEqual(show.get_stack_frames(['testtest']), ['testtest'])

  def test_bad_cast(self):
    """Test getting bad-cast."""
    lines = [
        'Bad-cast to content::RenderWidgetHostViewChildFrame1',
        'Bad-cast to content::RenderWidgetHostViewChildFrame1 from',
        ('Bad-cast to content::RenderWidgetHostViewChildFrame1 from'
         ' content::RenderWidgetHostViewAura'),
        ('Bad-cast to base::IsValueInRangeForNumericType<int>('
         ' std::floor(rect.x() * x_scale)) from invalid ptr')
    ]
    self.assertEqual(
        show.get_stack_frames(lines), [
            'content::RenderWidgetHostViewChildFrame1',
            'content::RenderWidgetHostViewChildFrame1',
            'content::RenderWidgetHostViewChildFrame1',
            ('base::IsValueInRangeForNumericType<int>( std::floor(rect.x()'
             ' * x_scale))'),
        ])

  def test_in(self):
    """Test getting frame in file."""
    lines = [
        'char_upper_api in file_path.cc',
        ('base::IsValueInRangeForNumericType<int>( std::floor(rect.x()'
         ' * x_scale)) in rect')
    ]
    self.assertEqual(
        show.get_stack_frames(lines), [
            'char_upper_api',
            ('base::IsValueInRangeForNumericType<int>( std::floor(rect.x()'
             ' * x_scale))')
        ])


class ConvertToLinesTest(unittest.TestCase):
  """Test convert_to_lines."""

  def test_empty(self):
    """Test empty trace."""
    self.assertEqual([], show.convert_to_lines('', [], ''))
    self.assertEqual([], show.convert_to_lines(' \n ', [], ''))

  def test_convert(self):
    """Test convert."""
    expected = [
        show.Line(1, 'a', False),
        show.Line(2, 'b', True),
        show.Line(3, 'c', True),
        show.Line(4, 'd', False),
        show.Line(5, 'e', False),
    ]
    self.assertEqual(expected,
                     show.convert_to_lines('a\nb\nc\nd\ne', ['b', 'c'], 'd'))

  def _test_special_type(self, crash_type):
    expected = [
        show.Line(1, 'a', False),
        show.Line(2, 'ERROR: something', True),
        show.Line(3, 'c', False),
        show.Line(4, 'SUMMARY: test', False),
        show.Line(5, 'e', False),
    ]
    self.assertEqual(
        expected,
        show.convert_to_lines('a\nERROR: something\nc\nSUMMARY: test\ne',
                              ['dontcare'], crash_type))

  def test_convert_out_of_memory(self):
    """Test convert out-of-memory."""
    self._test_special_type('Out-of-memory something')

  def test_convert_timeout(self):
    """Test convert timeout."""
    self._test_special_type('Timeout something')

  def test_convert_v8_correctness(self):
    """Test convert V8 correctness failure."""
    self._test_special_type('V8 correctness failure something')


class HighlightCommonStackFramesTest(unittest.TestCase):
  """Test highlight_common_stack_frames."""

  def test_highlight(self):
    """Ensure it highlights the last 3 lines of the first two crash stacks."""
    stack = '\n'.join([
        'random', '#0 0x3 test0', '#1 0x2 test2', '#2 0x1 test1',
        '#3 0x0 test0', 'random', '#0 0x4 test4', '#1 0x2 test2',
        '#2 0x1 test1', '#3 0x0 test0', 'random', '#0 0x5 test5',
        '#1 0x2 test2', '#2 0x1 test1', '#3 0x0 test0'
    ])
    expected = '\n'.join([
        'random', '#0 0x3 test0', '<b>#1 0x2 test2</b>', '<b>#2 0x1 test1</b>',
        '<b>#3 0x0 test0</b>', 'random', '#0 0x4 test4', '<b>#1 0x2 test2</b>',
        '<b>#2 0x1 test1</b>', '<b>#3 0x0 test0</b>', 'random', '#0 0x5 test5',
        '#1 0x2 test2', '#2 0x1 test1', '#3 0x0 test0'
    ])
    self.assertEqual(show.highlight_common_stack_frames(stack), expected)

  def test_one_stack_frame(self):
    """Ensure it does nothing when there's only one stack frame"""
    stack = '\n'.join([
        'random', '#0 0x3 test0', '#1 0x2 test2', '#2 0x1 test1',
        '#3 0x0 test0', 'random'
    ])
    self.assertEqual(show.highlight_common_stack_frames(stack), stack)


class FilterStacktraceTest(unittest.TestCase):
  """Test filter_stacktrace."""

  def setUp(self):
    test_helpers.patch(self, [
        'handlers.testcase_detail.show.highlight_common_stack_frames',
    ])

    def highlight(stacktrace):
      return stacktrace

    self.mock.highlight_common_stack_frames.side_effect = highlight

  def test_xss(self):
    """Ensure that we escape untrusted stacktrace."""
    stack = 'aaaa\n<script>alert("XSS")</script>\ncccc'
    expected = 'aaaa\n&lt;script&gt;alert(&quot;XSS&quot;)&lt;/script&gt;\ncccc'
    self.assertEqual(
        show.filter_stacktrace(stack, 'type', {}, '', ''), expected)

  def test_asan_chromium(self):
    """Ensure it linkifies asan trace for chromium."""
    revisions_dict = {
        '/src': {
            'url': 'https://chromium.googlesource.com/chromium/src.git',
            'rev': '1d783bc2a3629b94c963debfa3feaee27092dd92',
        },
        'src/v8': {
            'url': 'https://chromium.googlesource.com/v8/v8.git',
            'rev': '7fb2c3b6db3f889ea95851ca11dcb731b07a7925',
        }
    }
    stack = '\n'.join(
        [r'#1 0xa6760 in Test /build/src/v8/src/api.cc:3', r'random'])

    expected = ('#1 0xa6760 in Test <a href="'
                'https://chromium.googlesource.com/v8/v8/+/'
                '7fb2c3b6db3f889ea95851ca11dcb731b07a7925/src/api.cc#3">'
                'v8/src/api.cc:3'
                '</a>\n'
                'random')
    self.assertEqual(
        show.filter_stacktrace(stack, 'type', revisions_dict, '', ''), expected)

  def test_asan_oss_fuzz(self):
    """Ensure it linkifies asan trace for oss-fuzz."""
    revisions_dict = {
        '/src/libass': {
            'url': 'https://github.com/libass/libass.git',
            'rev': '35dc4dd0e14e3afb4a2c7e319a3f4110e20c7cf2',
            'type': 'git'
        },
        '/src/fribidi': {
            'url': 'https://github.com/behdad/fribidi.git',
            'rev': '881b8d891cc61989ab8811b74d0e721f72bf913b',
            'type': 'git'
        }
    }
    stack = '\n'.join(
        [r'#1 0xa6760 in Test /src/fribidi/lib/common.h:3', r'random'])

    expected = ('#1 0xa6760 in Test <a href="'
                'https://github.com/behdad/fribidi/blob/'
                '881b8d891cc61989ab8811b74d0e721f72bf913b/lib/common.h#L3">'
                'fribidi/lib/common.h:3'
                '</a>\n'
                'random')
    self.assertEqual(
        show.filter_stacktrace(stack, 'type', revisions_dict, '', ''), expected)

  def test_asan_v8(self):
    """Ensure it linkifies v8 win trace for chromium."""
    revisions_dict = {
        '/src': {
            'url': 'https://chromium.googlesource.com/v8/v8.git',
            'rev': '7fb2c3b6db3f889ea95851ca11dcb731b07a7925',
        }
    }
    stack = '\n'.join([(r'v8::internal::RootVisitor::VisitRootPointer '
                        r'[0x011517C5+53] (C:\b\c\b\win_asan_release\src\v8'
                        r'\src\visitors.h:69)'), r'random'])
    expected = ('v8::internal::RootVisitor::VisitRootPointer [0x011517C5+53] '
                '(<a href="https://chromium.googlesource.com/v8/v8/+/'
                '7fb2c3b6db3f889ea95851ca11dcb731b07a7925/v8/src/visitors.h#69"'
                '>v8/src/visitors.h:69</a>)\n'
                'random')

    self.assertEqual(
        show.filter_stacktrace(stack, 'type', revisions_dict, '', ''), expected)

  def test_no_linkify(self):
    """Ensure that we don't linkify a non-stack frame line."""
    revisions_dict = {
        '/src': {
            'url': 'https://chromium.googlesource.com/chromium/src.git',
            'rev': '1d783bc2a3629b94c963debfa3feaee27092dd92',
        },
        'src/v8': {
            'url': 'https://chromium.googlesource.com/v8/v8.git',
            'rev': '7fb2c3b6db3f889ea95851ca11dcb731b07a7925',
        }
    }
    stack = '\n'.join([
        ('../../net/spdy/chromium/spdy_stream.cc:227:21: runtime error: '
         'signed integer overflow:159714659 + 1996488831 cannot be represented '
         'in type \'int\''), 'random'
    ])
    expected = '\n'.join([
        ('../../net/spdy/chromium/spdy_stream.cc:227:21: runtime error: '
         'signed integer overflow:159714659 + 1996488831 cannot be represented '
         'in type &#x27;int&#x27;'), 'random'
    ])

    self.assertEqual(
        show.filter_stacktrace(stack, 'type', revisions_dict, '', ''), expected)

  def test_linkify_if_needed(self):
    """Ensure that we don't escape prelinked kernel links."""
    stack = ('[<ffffff90082bf4bc>] SyS_fchownat+0xdc/0x168 http://go/pakernel/'
             'msm-google/+/40e9b2ff3a280a8775cfcd5841e530ce78f94355/fs/open.c'
             '#622;msm-google/fs/open.c;')
    expected_stack = ('[&lt;ffffff90082bf4bc&gt;] SyS_fchownat+0xdc/0x168 '
                      '<a href="http://go/pakernel/'
                      'msm-google/+/'
                      '40e9b2ff3a280a8775cfcd5841e530ce78f94355/fs/open.c#622">'
                      'msm-google/fs/open.c</a>')
    self.assertEqual(
        show.filter_stacktrace(stack, 'type', {}, 'android', ''),
        expected_stack)


@test_utils.with_cloud_emulators('datastore')
class GetTestcaseTest(unittest.TestCase):
  """Test get_testcase."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.revisions.get_component_range_list',
        'clusterfuzz._internal.build_management.revisions.get_component_revisions_dict',
        'clusterfuzz._internal.config.db_config.get',
        'clusterfuzz._internal.config.db_config.get_value',
        'clusterfuzz._internal.config.db_config.get_value_for_job',
        'clusterfuzz._internal.datastore.data_handler.get_stacktrace',
        'handlers.testcase_detail.show.filter_stacktrace',
        'libs.issue_management.issue_tracker_utils.get_issue_url',
        'libs.access.can_user_access_testcase',
        'libs.access.has_access',
        'libs.auth.is_current_user_admin',
        'libs.form.generate_csrf_token',
        'libs.helpers.get_user_email',
        'clusterfuzz._internal.metrics.crash_stats.get_last_crash_time',
    ])

    self.mock.has_access.return_value = False
    self.mock.get_user_email.return_value = 'test@test.com'

    self.mock.get_component_range_list.return_value = [{
        'component': 'name',
        'link_text': '0:revision'
    }]

    self.mock.get_component_revisions_dict.return_value = {
        '/src': {
            'name': 'name',
            'rev': 'revision'
        }
    }

    self.mock.is_current_user_admin.return_value = False

    self.make_token().put()
    os.environ['ISSUE_TRACKER'] = 'test-issue-tracker'

  def make_token(self):
    token = data_types.CSRFToken()
    token.user_email = self.mock.get_user_email.return_value
    return token

  def test_no_testcase_id(self):
    """Test no testcase id."""
    with self.assertRaises(helpers.EarlyExitException) as cm:
      show.get_testcase_detail_by_id(None)

    self.assertEqual(cm.exception.status, 404)
    self.assertEqual(str(cm.exception), 'No test case specified!')

  def test_no_testcase(self):
    """Test invalid testcase."""
    with self.assertRaises(helpers.EarlyExitException) as cm:
      show.get_testcase_detail_by_id(1)

    self.assertEqual(cm.exception.status, 404)
    self.assertEqual(str(cm.exception), 'Invalid test case!')

  def test_forbidden(self):
    """Test forbidden testcase."""
    self.mock.can_user_access_testcase.return_value = False
    data_types.Testcase().put()

    with self.assertRaises(helpers.EarlyExitException) as cm:
      show.get_testcase_detail_by_id(2)

    self.assertEqual(cm.exception.status, 403)
    self.assertEqual(str(cm.exception), '')

  def test_reproducible_get(self):
    """Test valid reproducible testcase."""
    testcase = data_types.Testcase()
    testcase.job_type = 'linux_asan_chrome'
    testcase.crash_type = 'crash_type1\ncrash_type2'
    testcase.crash_address = 'crash_address'
    testcase.crash_state = 'crash_state'
    testcase.crash_revision = 123
    testcase.regression = None
    testcase.fixed = None
    testcase.fuzzed_keys = None
    testcase.minimized_keys = None
    testcase.timestamp = datetime.datetime(1970, 1, 1)
    testcase.project_name = 'chromium'
    testcase.one_time_crasher_flag = False
    testcase.fuzzer_name = 'fuzzer1'
    testcase.put()

    job = data_types.Job()
    job.name = 'linux_asan_chrome'
    job.custom_binary_revision = 1234
    job.environment_string = 'HELP_URL=help_url'
    job.put()

    self.mock.can_user_access_testcase.return_value = True
    self.mock.get_issue_url.return_value = 'issue_url'
    self.mock.get_stacktrace.return_value = 'crash_stacktrace'
    self.mock.filter_stacktrace.return_value = 'crash_stacktrace'
    self.mock.generate_csrf_token.return_value = 'csrf'

    result = show.get_testcase_detail_by_id(2)
    expected_subset = {
        'id':
            2,
        'crash_type':
            'crash_type1 crash_type2',
        'crash_address':
            'crash_address',
        'crash_state':
            'crash_state',
        'crash_state_lines': ['crash_state'],
        'crash_revision':
            123,
        'csrf_token':
            'csrf',
        'external_user':
            True,
        'footer':
            '',
        'fixed':
            'NO',
        'issue_url':
            'issue_url',
        'metadata': {},
        'minimized_testcase_size':
            None,
        'needs_refresh':
            True,
        'original_testcase_size':
            None,
        'privileged_user':
            False,
        'regression':
            'Pending',
        'security_severity':
            None,
        'show_impact':
            True,
        'show_blame':
            True,
        'auto_delete_timestamp':
            None,
        'auto_close_timestamp':
            None,
        'memory_tool_display_label':
            'Sanitizer',
        'memory_tool_display_value':
            'address (ASAN)',
        'last_tested':
            'name: 0:revision<br />',
        'is_admin_or_not_oss_fuzz':
            True,
        'has_issue_tracker':
            True,
        'reproduction_help_url':
            'help_url',
        'fuzzer_display':
            collections.OrderedDict([('engine', None), ('target', None),
                                     ('name', 'fuzzer1'),
                                     ('fully_qualified_name', 'fuzzer1')]),
    }

    self.maxDiff = None  # pylint: disable=invalid-name
    self.assertDictContainsSubset(expected_subset, result)
    self.assertEqual(result['testcase'].key.id(), testcase.key.id())

    self.assertDictContainsSubset({
        'lines': [show.Line(1, 'crash_stacktrace', False)]
    }, result['crash_stacktrace'])
    self.assertDictContainsSubset({
        'lines': [show.Line(1, 'crash_stacktrace', False)]
    }, result['last_tested_crash_stacktrace'])

  def test_unreproducible_get(self):
    """Test valid unreproducible testcase."""
    self.mock.get_last_crash_time.return_value = datetime.datetime(2000, 1, 1)

    testcase = data_types.Testcase()
    testcase.job_type = 'windows_asan_chrome'
    testcase.crash_type = 'crash_type1\ncrash_type2'
    testcase.crash_address = 'crash_address'
    testcase.crash_state = 'crash_state'
    testcase.crash_revision = 123
    testcase.regression = None
    testcase.fixed = None
    testcase.fuzzed_keys = None
    testcase.minimized_keys = None
    testcase.timestamp = datetime.datetime(1970, 1, 1)
    testcase.project_name = 'chromium'
    testcase.one_time_crasher_flag = True
    testcase.fuzzer_name = 'libFuzzer'
    testcase.overridden_fuzzer_name = 'libFuzzer_test_fuzzer'
    testcase.put()
    testcase.set_metadata('fuzzer_binary_name', 'test_fuzzer')

    job = data_types.Job()
    job.name = 'windows_asan_chrome'
    job.custom_binary_revision = 1234
    job.environment_string = 'HELP_URL=help_url'
    job.put()

    self.mock.can_user_access_testcase.return_value = True
    self.mock.get_issue_url.return_value = 'issue_url'
    self.mock.get_stacktrace.return_value = 'crash_stacktrace'
    self.mock.filter_stacktrace.return_value = 'crash_stacktrace'
    self.mock.generate_csrf_token.return_value = 'csrf'

    result = show.get_testcase_detail_by_id(2)
    expected_subset = {
        'id':
            2,
        'crash_type':
            'crash_type1 crash_type2',
        'crash_address':
            'crash_address',
        'crash_state':
            'crash_state',
        'crash_state_lines': ['crash_state'],
        'crash_revision':
            123,
        'csrf_token':
            'csrf',
        'external_user':
            True,
        'footer':
            '',
        'fixed':
            'NO',
        'issue_url':
            'issue_url',
        'metadata': {
            'fuzzer_binary_name': 'test_fuzzer'
        },
        'minimized_testcase_size':
            None,
        'needs_refresh':
            True,
        'original_testcase_size':
            None,
        'privileged_user':
            False,
        'regression':
            'Pending',
        'security_severity':
            None,
        'show_impact':
            False,
        'show_blame':
            True,
        'auto_delete_timestamp':
            947289600.0,
        'auto_close_timestamp':
            None,
        'memory_tool_display_label':
            'Sanitizer',
        'memory_tool_display_value':
            'address (ASAN)',
        'last_tested':
            'name: 0:revision<br />',
        'is_admin_or_not_oss_fuzz':
            True,
        'has_issue_tracker':
            True,
        'reproduction_help_url':
            'help_url',
        'fuzzer_display':
            collections.OrderedDict(
                [('engine', 'libFuzzer'), ('target', 'test_fuzzer'),
                 ('name', 'libFuzzer'), ('fully_qualified_name',
                                         'libFuzzer_test_fuzzer')]),
    }

    self.maxDiff = None  # pylint: disable=invalid-name
    self.assertDictContainsSubset(expected_subset, result)
    self.assertEqual(result['testcase'].key.id(), testcase.key.id())

    self.assertDictContainsSubset({
        'lines': [show.Line(1, 'crash_stacktrace', False)]
    }, result['crash_stacktrace'])
    self.assertDictContainsSubset({
        'lines': [show.Line(1, 'crash_stacktrace', False)]
    }, result['last_tested_crash_stacktrace'])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Deploy tests."""
# pylint: disable=protected-access
import datetime
import json
import os
import sys
import unittest

import mock
from pyfakefs import fake_filesystem_unittest
import yaml

from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from local.butler import deploy


@mock.patch('local.butler.deploy.RETRY_WAIT_SECONDS', 0)
class DeployTest(fake_filesystem_unittest.TestCase):
  """Deploy tests."""

  def setUp(self):
    """Setup for deploy test."""
    real_cwd = os.path.realpath(os.getcwd())
    test_utils.set_up_pyfakefs(self)
    self.fs.add_real_directory(
        os.path.join(real_cwd, 'src', 'appengine'), read_only=False)

    helpers.patch_environ(self)
    helpers.patch(self, [
        'local.butler.common.execute',
        'local.butler.common.Gcloud.run',
        'local.butler.common.has_file_in_path',
        'local.butler.deploy.now',
        'os.remove',
    ])
    self.mock.execute.side_effect = self._mock_execute
    self.mock.has_file_in_path.return_value = True
    self.deploy_failure_count = 0

    os.environ['ROOT_DIR'] = '.'
    self.mock.now.return_value = datetime.datetime(2017, 1, 3, 12, 1)
    self.manifest_target = 'clusterfuzz-source.manifest'
    if sys.version_info.major == 3:
      self.manifest_target += '.3'

  def _check_env_variables(self, yaml_paths):
    """Check that environment variables are written to yaml paths."""
    for yaml_path in yaml_paths:
      with open(yaml_path) as f:
        data = yaml.safe_load(f)

      self.assertIn('env_variables', data)
      env_variables = data['env_variables']
      self.assertEqual('test-clusterfuzz', env_variables['APPLICATION_ID'])
      self.assertEqual('test-project', env_variables['PROJECT_NAME'])
      self.assertEqual('test-corpus-bucket', env_variables['CORPUS_BUCKET'])
      self.assertEqual('test-quarantine-bucket',
                       env_variables['QUARANTINE_BUCKET'])
      self.assertEqual('test-shared-corpus-bucket',
                       env_variables['SHARED_CORPUS_BUCKET'])

  def _check_no_env_variables(self, yaml_paths):
    """Check that environment variables are not written to yaml paths."""
    for yaml_path in yaml_paths:
      with open(yaml_path) as f:
        data = yaml.safe_load(f)

      self.assertNotIn('env_variables', data)

  # pylint: disable=unused-argument
  def _mock_execute(self, command, *args, **kwargs):
    """Mock execute."""
    if 'app deploy' in command:
      if self.deploy_failure_count == 0:
        return (0, b'ok')

      self.deploy_failure_count -= 1
      return (1, b'failure')

    if 'app describe' in command:
      return (0, b'us-central')

    if 'describe redis-instance' in command:
      return (0, b'redis-ip')

    if 'describe' in command:
      return (1, b'')

    if 'versions list' in command:
      return (0,
              json.dumps([
                  {
                      'id': 'v1',
                      'last_deployed_time': {
                          'year': 2017,
                          'month': 1,
                          'day': 2,
                          'hour': 0,
                          'minute': 0,
                          'second': 0,
                      },
                      'traffic_split': 0.0,
                  },
                  {
                      'id': 'v2',
                      'last_deployed_time': {
                          'year': 2017,
                          'month': 1,
                          'day': 3,
                          'hour': 0,
                          'minute': 0,
                          'second': 0,
                      },
                      'traffic_split': 0.0,
                  },
                  {
                      'id': 'current',
                      'last_deployed_time': {
                          'year': 2017,
                          'month': 1,
                          'day': 3,
                          'hour': 12,
                          'minute': 0,
                          'second': 1,
                      },
                      'traffic_split': 1.0,
                  },
              ]).encode())

    return (0, b'')

  def test_app(self):
    """Test deploy app."""
    deploy._prod_deployment_helper('/config_dir',
                                   ['/windows.zip', '/mac.zip', '/linux.zip'])

    self.mock.run.assert_has_calls([
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'describe',
                  'pubsub'),
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'update',
                  'pubsub', '--config=./configs/test/pubsub/queues.yaml'),
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'describe',
                  'bigquery'),
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'update',
                  'bigquery', '--config=./configs/test/bigquery/datasets.yaml'),
    ])

    self.mock.execute.assert_has_calls([
        mock.call(
            'gcloud app deploy --no-stop-previous-version --quiet '
            '--project=test-clusterfuzz  '
            'src/appengine/index.yaml '
            'src/appengine/app.yaml '
            'src/appengine/cron.yaml '
            'src/appengine/cron-service.yaml',
            exit_on_error=False),
        mock.call('gcloud app versions list --format=json '
                  '--project=test-clusterfuzz --service=default'),
        mock.call(
            'gcloud app versions delete --quiet --project=test-clusterfuzz '
            '--service=default v1'),
        mock.call('gcloud app versions list --format=json '
                  '--project=test-clusterfuzz --service=cron-service'),
        mock.call(
            'gcloud app versions delete --quiet --project=test-clusterfuzz '
            '--service=cron-service v1'),
        mock.call('gsutil cp /windows.zip gs://test-deployment-bucket/'
                  'windows.zip'),
        mock.call('gsutil cp /mac.zip gs://test-deployment-bucket/'
                  'mac.zip'),
        mock.call('gsutil cp /linux.zip gs://test-deployment-bucket/'
                  'linux.zip'),
        mock.call('gsutil cp -a public-read src/appengine/resources/'
                  'clusterfuzz-source.manifest '
                  'gs://test-deployment-bucket/' + self.manifest_target),
        mock.call('python butler.py run setup --config-dir /config_dir '
                  '--non-dry-run'),
    ])
    self._check_env_variables([
        'src/appengine/app.yaml',
        'src/appengine/cron-service.yaml',
    ])
    self._check_no_env_variables(
        ['src/appengine/cron.yaml', 'src/appengine/index.yaml'])

  def test_app_staging(self):
    """Test deploy app to staging."""
    deploy._staging_deployment_helper()

    self.mock.execute.assert_has_calls([
        mock.call(
            'gcloud app deploy --stop-previous-version --quiet '
            '--project=test-clusterfuzz  '
            'src/appengine/staging.yaml',
            exit_on_error=False),
        mock.call('gcloud app versions list --format=json '
                  '--project=test-clusterfuzz --service=staging'),
        mock.call(
            'gcloud app versions delete --quiet --project=test-clusterfuzz '
            '--service=staging v1 v2'),
    ])
    self._check_env_variables(['src/appengine/staging.yaml'])

  def test_app_retry(self):
    """Test deploy app with retries."""
    self.deploy_failure_count = 1

    deploy._prod_deployment_helper('/config_dir',
                                   ['/windows.zip', '/mac.zip', '/linux.zip'])

    self.mock.run.assert_has_calls([
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'describe',
                  'pubsub'),
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'update',
                  'pubsub', '--config=./configs/test/pubsub/queues.yaml'),
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'describe',
                  'bigquery'),
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'update',
                  'bigquery', '--config=./configs/test/bigquery/datasets.yaml'),
    ])

    self.mock.execute.assert_has_calls([
        mock.call(
            'gcloud app deploy --no-stop-previous-version --quiet '
            '--project=test-clusterfuzz  '
            'src/appengine/index.yaml '
            'src/appengine/app.yaml '
            'src/appengine/cron.yaml '
            'src/appengine/cron-service.yaml',
            exit_on_error=False),
        mock.call(
            'gcloud app deploy --no-stop-previous-version --quiet '
            '--project=test-clusterfuzz  '
            'src/appengine/index.yaml '
            'src/appengine/app.yaml '
            'src/appengine/cron.yaml '
            'src/appengine/cron-service.yaml',
            exit_on_error=False),
        mock.call('gcloud app versions list --format=json '
                  '--project=test-clusterfuzz --service=default'),
        mock.call(
            'gcloud app versions delete --quiet --project=test-clusterfuzz '
            '--service=default v1'),
        mock.call('gcloud app versions list --format=json '
                  '--project=test-clusterfuzz --service=cron-service'),
        mock.call(
            'gcloud app versions delete --quiet --project=test-clusterfuzz '
            '--service=cron-service v1'),
        mock.call('gsutil cp /windows.zip gs://test-deployment-bucket/'
                  'windows.zip'),
        mock.call('gsutil cp /mac.zip gs://test-deployment-bucket/'
                  'mac.zip'),
        mock.call('gsutil cp /linux.zip gs://test-deployment-bucket/'
                  'linux.zip'),
        mock.call('gsutil cp -a public-read src/appengine/resources/'
                  'clusterfuzz-source.manifest '
                  'gs://test-deployment-bucket/' + self.manifest_target),
        mock.call('python butler.py run setup --config-dir /config_dir '
                  '--non-dry-run'),
    ])
    self._check_env_variables([
        'src/appengine/app.yaml',
        'src/appengine/cron-service.yaml',
    ])
    self._check_no_env_variables(
        ['src/appengine/cron.yaml', 'src/appengine/index.yaml'])

  def test_app_retry_failure(self):
    """Test deploy app with retries (failure)."""
    self.deploy_failure_count = 4

    with self.assertRaises(SystemExit):
      deploy._prod_deployment_helper('/config_dir',
                                     ['/windows.zip', '/mac.zip', '/linux.zip'])

    self.mock.run.assert_has_calls([
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'describe',
                  'pubsub'),
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'update',
                  'pubsub', '--config=./configs/test/pubsub/queues.yaml'),
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'describe',
                  'bigquery'),
        mock.call(mock.ANY, 'deployment-manager', 'deployments', 'update',
                  'bigquery', '--config=./configs/test/bigquery/datasets.yaml'),
    ])

    self.mock.execute.assert_has_calls([
        mock.call(
            'gcloud app deploy --no-stop-previous-version --quiet '
            '--project=test-clusterfuzz  '
            'src/appengine/index.yaml '
            'src/appengine/app.yaml '
            'src/appengine/cron.yaml '
            'src/appengine/cron-service.yaml',
            exit_on_error=False),
        mock.call(
            'gcloud app deploy --no-stop-previous-version --quiet '
            '--project=test-clusterfuzz  '
            'src/appengine/index.yaml '
            'src/appengine/app.yaml '
            'src/appengine/cron.yaml '
            'src/appengine/cron-service.yaml',
            exit_on_error=False),
        mock.call(
            'gcloud app deploy --no-stop-previous-version --quiet '
            '--project=test-clusterfuzz  '
            'src/appengine/index.yaml '
            'src/appengine/app.yaml '
            'src/appengine/cron.yaml '
            'src/appengine/cron-service.yaml',
            exit_on_error=False),
        mock.call(
            'gcloud app deploy --no-stop-previous-version --quiet '
            '--project=test-clusterfuzz  '
            'src/appengine/index.yaml '
            'src/appengine/app.yaml '
            'src/appengine/cron.yaml '
            'src/appengine/cron-service.yaml',
            exit_on_error=False),
    ])


class FindFileExceedingLimitTest(fake_filesystem_unittest.TestCase):
  """Test finding files exceeding limit."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    self.fs.create_file('/test/small1', contents='aaa')
    self.fs.create_file('/test/small2', contents='aaa')
    self.fs.create_file('/test/dir1/small3', contents='aaa')
    self.fs.create_file('/test/dir1/small4', contents='aaa')
    self.fs.create_file('/test/dir1/dir1/small5', contents='aaa')
    self.fs.create_file('/test/dir2/small6', contents='aaa')

  def test_get_too_large_file(self):
    """Test getting a too large file."""
    self.fs.create_file('/test/dir1/dir1/too_large', contents='aaaaaa')
    self.assertEqual('/test/dir1/dir1/too_large',
                     deploy.find_file_exceeding_limit('/test', 5))

  def test_get_none(self):
    """Test when there's no too large file."""
    self.assertIsNone(deploy.find_file_exceeding_limit('/test', 10))


class GetRemoteShaTest(unittest.TestCase):
  """Test get_remote_sha."""

  def setUp(self):
    helpers.patch(self, ['local.butler.common.execute'])

  def test_get(self):
    """Test get_remote_sha."""
    self.mock.execute.return_value = (
        0, b'cbb7f93c7ddc1c3a3c98f45ebf5c3490a0c38e95        refs/heads/master')

    self.assertEqual(b'cbb7f93c7ddc1c3a3c98f45ebf5c3490a0c38e95',
                     deploy.get_remote_sha())


class IsDiffOriginMasterTest(unittest.TestCase):
  """Test is_diff_origin_master."""

  def setUp(self):
    helpers.patch(
        self,
        ['local.butler.common.execute', 'local.butler.deploy.get_remote_sha'])

    self.head = ''
    self.diff = ''

    def execute(cmd):
      if cmd == 'git fetch':
        return (0, '')
      if cmd == 'git rev-parse HEAD':
        return (0, self.head)
      if cmd == 'git diff origin/master --stat':
        return (0, self.diff)
      raise Exception()

    self.mock.execute.side_effect = execute

  def test_good(self):
    """Test good."""
    self.diff = ''
    self.mock.get_remote_sha.return_value = 'sha'
    self.head = 'sha'

    self.assertFalse(deploy.is_diff_origin_master())

  def test_diff(self):
    """Test diff."""
    self.diff = 'something'
    self.mock.get_remote_sha.return_value = 'sha'
    self.head = 'sha'

    self.assertTrue(deploy.is_diff_origin_master())

  def test_diff_sha(self):
    """Test different sha."""
    self.diff = ''
    self.mock.get_remote_sha.return_value = 'sha'
    self.head = 'sha2'

    self.assertTrue(deploy.is_diff_origin_master())


class VersionsToDeleteTest(unittest.TestCase):
  """Test _versions_to_delete."""

  def setUp(self):
    helpers.patch(self, [
        'local.butler.deploy.now',
    ])

  def test_single_version(self):
    """Test single revision."""
    self.mock.now.return_value = datetime.datetime(2017, 1, 1, 0, 0)
    to_delete = deploy._versions_to_delete([
        deploy.Version('1', datetime.datetime(2017, 1, 1, 0, 0), 1.0),
    ], 24 * 60)

    self.assertEqual([], to_delete)

  def test_two_revisions(self):
    """Test two revision."""
    self.mock.now.return_value = datetime.datetime(2017, 1, 1, 0, 0)
    to_delete = deploy._versions_to_delete([
        deploy.Version('1', datetime.datetime(2016, 1, 1, 0, 0), 0.0),
        deploy.Version('2', datetime.datetime(2017, 1, 1, 0, 0), 1.0),
    ], 24 * 60)

    self.assertEqual([], to_delete)

  def test_cutoff(self):
    """Test various cutoffs."""
    self.mock.now.return_value = datetime.datetime(2017, 1, 30, 0, 0)
    to_delete = deploy._versions_to_delete([
        deploy.Version('1', datetime.datetime(2016, 1, 1, 0, 0), 0.0),
        deploy.Version('2', datetime.datetime(2017, 1, 28, 23, 59), 0.0),
        deploy.Version('3', datetime.datetime(2017, 1, 29, 0, 0), 0.0),
        deploy.Version('4', datetime.datetime(2017, 1, 30, 0, 0), 1.0),
    ], 24 * 60)

    self.assertEqual([
        deploy.Version('1', datetime.datetime(2016, 1, 1, 0, 0), 0.0),
        deploy.Version('2', datetime.datetime(2017, 1, 28, 23, 59), 0.0),
    ], to_delete)

    to_delete = deploy._versions_to_delete([
        deploy.Version('1', datetime.datetime(2016, 1, 1, 0, 0), 0.0),
        deploy.Version('2', datetime.datetime(2017, 1, 28, 23, 59), 0.0),
        deploy.Version('3', datetime.datetime(2017, 1, 29, 0, 1), 0.0),
        deploy.Version('4', datetime.datetime(2017, 1, 30, 0, 0), 1.0),
    ], 24 * 60)

    self.assertEqual([
        deploy.Version('1', datetime.datetime(2016, 1, 1, 0, 0), 0.0),
    ], to_delete)

    to_delete = deploy._versions_to_delete([
        deploy.Version('1', datetime.datetime(2016, 1, 1, 0, 0), 0.0),
        deploy.Version('2', datetime.datetime(2017, 1, 29, 0, 1), 0.0),
        deploy.Version('3', datetime.datetime(2017, 1, 29, 0, 2), 0.0),
        deploy.Version('4', datetime.datetime(2017, 1, 30, 0, 0), 1.0),
    ], 24 * 60)

    self.assertEqual([], to_delete)

    # Latest version should never be deleted.
    to_delete = deploy._versions_to_delete([
        deploy.Version('1', datetime.datetime(2016, 1, 1, 0, 0), 0.0),
        deploy.Version('2', datetime.datetime(2016, 1, 29, 0, 1), 0.0),
        deploy.Version('3', datetime.datetime(2016, 1, 29, 0, 2), 0.0),
        deploy.Version('4', datetime.datetime(2016, 1, 30, 0, 0), 1.0),
    ], 24 * 60)

    self.assertEqual([
        deploy.Version('1', datetime.datetime(2016, 1, 1, 0, 0), 0.0),
        deploy.Version('2', datetime.datetime(2016, 1, 29, 0, 1), 0.0),
        deploy.Version('3', datetime.datetime(2016, 1, 29, 0, 2), 0.0),
    ], to_delete)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Query handles constructing BigQuery's SQL."""

import json

from libs.query import base


class Query(base.Query):
  """Represent a query for BigQuery's query. Named parameter mode is used."""

  def __init__(self):
    self.conditions = []
    self.or_groups = []
    self.sort_by = None

  def raw_filter(self, cond):
    """Add raw filter directly."""
    self.conditions.append(cond)

  def filter(self, field, value, operator='='):
    """Filter by a single value."""
    # json.dumps converts Python literals to BigQuery literals perfectly well.
    # See tests.
    self.conditions.append('%s %s %s' % (field, operator, json.dumps(value)))

  def filter_in(self, field, values):
    """Filter by multiple values."""
    literals = [json.dumps(v) for v in values]
    self.conditions.append('%s IN (%s)' % (field, ', '.join(literals)))

  def union(self, *queries):
    """Combine queries with OR conditions."""
    self.or_groups.append(queries)

  def new_subquery(self):
    """Generate a new subquery."""
    return Query()

  def get_where_clause(self):
    """Get the where clause."""
    subquery_wheres = []
    for or_queries in self.or_groups:
      or_cond = ' OR '.join([sub.get_where_clause() for sub in or_queries])
      subquery_wheres.append('(%s)' % or_cond)

    all_conds = self.conditions + subquery_wheres

    if not all_conds:
      return ''

    return '(%s)' % ' AND '.join(all_conds)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""honggfuzz engine interface."""

import glob
import os
import re

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import dictionary_manager
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import new_process
from clusterfuzz.fuzz import engine

_CLEAN_EXIT_SECS = 10
_RSS_LIMIT = 2560
_TIMEOUT = 25
_DEFAULT_ARGUMENTS = [
    '-n',
    '1',  # single threaded
    '--exit_upon_crash',
    '-v',  # output to stderr
    '-z',  # use clang instrumentation
    '-P',  # persistent mode
    '-S',  # enable sanitizers
    '--rlimit_rss',
    str(_RSS_LIMIT),
    '--timeout',
    str(_TIMEOUT),
]

_CRASH_REGEX = re.compile('Crash: saved as \'(.*)\'')
_HF_SANITIZER_LOG_PREFIX = 'HF.sanitizer.log'
_STATS_PREFIX = 'Summary '


class HonggfuzzError(Exception):
  """Base exception class."""


def _get_runner():
  """Get the honggfuzz runner."""
  honggfuzz_path = os.path.join(environment.get_value('BUILD_DIR'), 'honggfuzz')
  if not os.path.exists(honggfuzz_path):
    raise HonggfuzzError('honggfuzz not found in build')

  os.chmod(honggfuzz_path, 0o755)
  return new_process.UnicodeProcessRunner(honggfuzz_path)


def _find_sanitizer_stacktrace(reproducers_dir):
  """Find the sanitizer stacktrace from the reproducers dir."""
  for stacktrace_path in glob.glob(
      os.path.join(reproducers_dir, _HF_SANITIZER_LOG_PREFIX + '*')):
    with open(stacktrace_path, 'rb') as f:
      return utils.decode_to_unicode(f.read())

  return None


def _get_reproducer_path(line):
  """Get the reproducer path, if any."""
  crash_match = _CRASH_REGEX.match(line)
  if not crash_match:
    return None

  return crash_match.group(1)


def _get_stats(line):
  """Get stats, if any."""
  if not line.startswith(_STATS_PREFIX):
    return None

  parts = line[len(_STATS_PREFIX):].split()
  stats = {}

  for part in parts:
    if ':' not in part:
      logs.log_error('Invalid stat part.', value=part)

    key, value = part.split(':', 2)
    try:
      stats[key] = int(value)
    except (ValueError, TypeError):
      logs.log_error('Invalid stat value.', key=key, value=value)

  return stats


class HonggfuzzEngine(engine.Engine):
  """honggfuzz engine implementation."""

  @property
  def name(self):
    return 'honggfuzz'

  def prepare(self, corpus_dir, target_path, build_dir):  # pylint: disable=unused-argument
    """Prepare for a fuzzing session, by generating options. Returns a
    FuzzOptions object.

    Args:
      corpus_dir: The main corpus directory.
      target_path: Path to the target.
      build_dir: Path to the build directory.

    Returns:
      A FuzzOptions object.
    """
    os.chmod(target_path, 0o775)
    arguments = []
    dict_path = dictionary_manager.get_default_dictionary_path(target_path)
    if os.path.exists(dict_path):
      arguments.extend(['--dict', dict_path])

    return engine.FuzzOptions(corpus_dir, arguments, {})

  def fuzz(self, target_path, options, reproducers_dir, max_time):
    """Run a fuzz session.

    Args:
      target_path: Path to the target.
      options: The FuzzOptions object returned by prepare().
      reproducers_dir: The directory to put reproducers in when crashes
          are found.
      max_time: Maximum allowed time for the fuzzing to run.

   Returns:
      A FuzzResult object.
    """
    runner = _get_runner()
    arguments = _DEFAULT_ARGUMENTS[:]
    arguments.extend(options.arguments)
    arguments.extend([
        '--input',
        options.corpus_dir,
        '--workspace',
        reproducers_dir,
        '--run_time',
        str(max_time),
        '--',
        target_path,
    ])

    fuzz_result = runner.run_and_wait(
        additional_args=arguments, timeout=max_time + _CLEAN_EXIT_SECS)
    log_lines = fuzz_result.output.splitlines()
    sanitizer_stacktrace = _find_sanitizer_stacktrace(reproducers_dir)

    crashes = []
    stats = None
    for line in log_lines:
      reproducer_path = _get_reproducer_path(line)
      if reproducer_path:
        crashes.append(
            engine.Crash(reproducer_path, sanitizer_stacktrace or '', [],
                         int(fuzz_result.time_executed)))
        continue

      stats = _get_stats(line)

    if stats is None:
      stats = {}

    return engine.FuzzResult(fuzz_result.output, fuzz_result.command, crashes,
                             stats, fuzz_result.time_executed)

  def reproduce(self, target_path, input_path, arguments, max_time):  # pylint: disable=unused-argument
    """Reproduce a crash given an input.

    Args:
      target_path: Path to the target.
      input_path: Path to the reproducer input.
      arguments: Additional arguments needed for reproduction.
      max_time: Maximum allowed time for the reproduction.

    Returns:
      A ReproduceResult.
    """
    os.chmod(target_path, 0o775)
    runner = new_process.UnicodeProcessRunner(target_path)
    with open(input_path) as f:
      result = runner.run_and_wait(timeout=max_time, stdin=f)

    return engine.ReproduceResult(result.command, result.return_code,
                                  result.time_executed, result.output)

  def minimize_corpus(self, target_path, arguments, input_dirs, output_dir,
                      reproducers_dir, max_time):
    """Optional (but recommended): run corpus minimization.

    Args:
      target_path: Path to the target.
      arguments: Additional arguments needed for corpus minimization.
      input_dirs: Input corpora.
      output_dir: Output directory to place minimized corpus.
      reproducers_dir: The directory to put reproducers in when crashes are
          found.
      max_time: Maximum allowed time for the minimization.

    Returns:
      A FuzzResult object.
    """
    # TODO(ochang): Implement this.
    raise NotImplementedError



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Linux Kernel Library specific modules."""
from clusterfuzz._internal.system import environment

from . import constants


def get_lkl_binary_name(unsymbolized_crash_stacktrace_split):
  """Returns the lkl binary name from a stack trace."""
  for line in unsymbolized_crash_stacktrace_split:
    match = constants.LINUX_KERNEL_LIBRARY_ASSERT_REGEX.match(line)
    if match:
      return match.group(1)

  return None


def is_lkl_stack_trace(unsymbolized_crash_stacktrace):
  """Is this an lkl stack trace?"""
  return (
      environment.is_lkl_job() and
      constants.LINUX_KERNEL_MODULE_STACK_TRACE in unsymbolized_crash_stacktrace
  )



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generic helper functions useful in tests (App Engine only)."""

import datetime

from clusterfuzz._internal.tests.test_libs import test_utils
from libs.issue_management import monorail
from libs.issue_management.monorail.comment import Comment
from libs.issue_management.monorail.issue import Issue


def create_generic_issue(created_days_ago=28):
  """Returns a simple issue object for use in tests."""
  issue = Issue()
  issue.cc = []
  issue.comment = ''
  issue.comments = []
  issue.components = []
  issue.labels = []
  issue.open = True
  issue.owner = 'owner@chromium.org'
  issue.status = 'Assigned'
  issue.id = 1
  issue.itm = create_issue_tracker_manager()

  # Test issue was created 1 week before the current (mocked) time.
  issue.created = (
      test_utils.CURRENT_TIME - datetime.timedelta(days=created_days_ago))

  return monorail.Issue(issue)


def create_generic_issue_comment(comment_body='Comment.',
                                 author='user@chromium.org',
                                 days_ago=21,
                                 labels=None):
  """Return a simple comment used for testing."""
  comment = Comment()
  comment.comment = comment_body
  comment.author = author
  comment.created = test_utils.CURRENT_TIME - datetime.timedelta(days=days_ago)
  comment.labels = labels

  if comment.labels is None:
    comment.labels = []

  return comment


def create_issue_tracker_manager():
  """Create a fake issue tracker manager."""

  class FakeIssueTrackerManager(object):
    """Mock issue tracker manager."""

    def __init__(self):
      self.project_name = 'test-project'
      self.issues = {}
      self.next_id = 1

    def get_issue(self, issue_id):
      """Get original issue."""
      issue = self.issues.get(issue_id)
      if not issue:
        return None

      issue.itm = self
      return issue

    def get_comments(self, issue):  # pylint: disable=unused-argument
      """Return an empty comment list."""
      return []

    def save(self, issue, *args, **kwargs):  # pylint: disable=unused-argument
      """Save an issue."""
      if issue.new:
        issue.id = self.next_id
        issue.new = False
        self.next_id += 1

      self.issues[issue.id] = issue

  return FakeIssueTrackerManager()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests fuzzers.mutator_plugin."""

import os
import shutil
import unittest

from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.bot.fuzzers import mutator_plugin
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class FindMutatorPluginTest(fake_filesystem_unittest.TestCase):
  """Tests find_mutator_plugin."""

  def setUp(self):
    helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)
    self.plugins_root_dir = '/plugins'
    os.environ['MUTATOR_PLUGINS_DIR'] = self.plugins_root_dir

  def test_find_mutator_plugin_with_usable(self):
    """Tests that the right path is returned by find_mutator_plugin when there
    is a usable mutator plugin available."""
    usable_plugin_path = os.path.join(
        self.plugins_root_dir, 'plugins',
        mutator_plugin.MUTATOR_SHARED_OBJECT_FILENAME)

    self.fs.create_file(usable_plugin_path)
    self.assertEqual(usable_plugin_path, mutator_plugin.find_mutator_plugin())

  def test_set_mutator_plugin_without_usable(self):
    """Tests that None is returned by find_mutator_plugin when there isn't a
    usable mutator plugin available."""
    self.assertIsNone(mutator_plugin.find_mutator_plugin())


# pylint: disable=protected-access
class GetDirectoryFunctionsTest(unittest.TestCase):
  """Tests functions for get plugin directories."""

  def setUp(self):
    helpers.patch_environ(self)
    self.plugins_root_dir = '/plugins'
    os.environ['MUTATOR_PLUGINS_DIR'] = self.plugins_root_dir

  def test_get_mutator_plugins_subdir(self):
    """Tests that _get_mutator_plugins_subdir returns the path to the correct
    subdirectory."""
    subdir = 'x'
    self.assertEqual(
        os.path.join(self.plugins_root_dir, subdir),
        mutator_plugin._get_mutator_plugins_subdir(subdir))

  def test_get_mutator_plugins_archives_dir(self):
    """Tests that _get_mutator_plugins_archives_dir returns the path to the
    mutator plugin archives directory."""
    self.assertEqual(
        os.path.join(self.plugins_root_dir,
                     mutator_plugin.ARCHIVES_SUBDIR_NAME),
        mutator_plugin._get_mutator_plugins_archives_dir())

  def test_get_mutator_plugins_unpacked_dir(self):
    """Tests that _get_mutator_plugins_unpacked_dir returns the path to the
    unpacked mutator plugin directory."""
    self.assertEqual(
        os.path.join(self.plugins_root_dir, mutator_plugin.PLUGINS_SUBDIR_NAME),
        mutator_plugin._get_mutator_plugins_unpacked_dir())


# pylint: disable=protected-access
class PluginGetterTest(fake_filesystem_unittest.TestCase):
  """Tests PluginGetter."""

  def setUp(self):
    """Setup for plugin getter test."""
    helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)
    os.environ['JOB_NAME'] = 'libfuzzer_asan_test'
    self.fuzzer_binary_name = 'test_fuzzer'
    self.name = 'myplugin'
    self.plugins_root_dir = '/plugins'
    os.environ['MUTATOR_PLUGINS_DIR'] = self.plugins_root_dir
    self.fs.create_dir(self.plugins_root_dir)
    self.plugin_getter = mutator_plugin.PluginGetter(self.fuzzer_binary_name)
    self.plugins_archives_dir = os.path.join(self.plugins_root_dir, 'archives')
    self.plugin_archive_filename = '%s-%s-%s.zip' % (
        self.name, os.environ['JOB_NAME'], self.fuzzer_binary_name)
    self.plugin_archive_path = os.path.join(self.plugins_archives_dir,
                                            self.plugin_archive_filename)
    self.plugins_dir = os.path.join(self.plugins_root_dir, 'plugins')

    helpers.patch(self, [
        'clusterfuzz._internal.google_cloud_utils.storage.copy_file_from',
        'clusterfuzz._internal.bot.fuzzers.mutator_plugin._get_mutator_plugins_from_bucket',
    ])

    def mocked_copy_file_from(gcs_url, file_path):
      expected_url = '%s/%s' % (mutator_plugin._get_mutator_plugins_bucket_url(
      ), self.plugin_archive_filename)

      self.assertEqual(expected_url, gcs_url)
      self.assertEqual(file_path, self.plugin_archive_path)
      return file_path

    self.mock.copy_file_from.side_effect = mocked_copy_file_from

  def test_create_directories(self):
    """Tests that create_directories creates the right directories."""
    shutil.rmtree(self.plugins_root_dir)
    self.fs.create_dir(self.plugins_root_dir)
    self.plugin_getter.create_directories()
    directories = [
        os.path.join(self.plugins_root_dir, 'plugins'),
        os.path.join(self.plugins_root_dir, 'archives')
    ]
    self.assertTrue(all(os.path.isdir(directory) for directory in directories))

  def test_recognizes_usable(self):
    """Tests that _is_plugin_usable recognizes a usable plugin archive."""
    self.assertTrue(
        self.plugin_getter._is_plugin_usable(self.plugin_archive_filename))

  def test_recognizes_unusable(self):
    """Tests that _is_plugin_usable recognizes an unusable plugin archive."""
    unusable_plugin_archive_filename = self.plugin_archive_filename.replace(
        self.fuzzer_binary_name, 'other_binary')
    self.assertFalse(
        self.plugin_getter._is_plugin_usable(unusable_plugin_archive_filename))

  def test_download_mutator_plugin_archive(self):
    """Tests that _download_mutator_plugin_archive downloads an archive to the
    correct location."""
    self.assertEqual(
        self.plugin_archive_path,
        mutator_plugin._download_mutator_plugin_archive(
            self.plugin_archive_filename))


class ExtractNameFromArchiveTest(unittest.TestCase):
  """Tests for _extract_name_from_archive."""

  def test_extract_name_from_archive(self):
    """Tests that _extract_name_from_archive extracts the name from the
    archive."""
    name = 'myplugin'
    fuzzer_binary_name = 'test_fuzzer'
    job_name = 'libfuzzer_asan_test'
    plugin_archive_filename = '%s-%s-%s.zip' % (name, job_name,
                                                fuzzer_binary_name)
    extracted_name, job_and_fuzz_target = (
        mutator_plugin._extract_name_from_archive(plugin_archive_filename))

    self.assertEqual(name, extracted_name)
    expected_job_and_fuzz_target = '%s-%s' % (job_name, fuzzer_binary_name)
    self.assertEqual(expected_job_and_fuzz_target, job_and_fuzz_target)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests testcase_manager."""
import datetime
import os
import shutil
import tempfile
import unittest

import mock
from pyfakefs import fake_filesystem_unittest

from clusterfuzz import stacktraces
from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.fuzzers.libFuzzer import \
    constants as libfuzzer_constants
from clusterfuzz._internal.bot.fuzzers.libFuzzer import \
    engine as libfuzzer_engine
from clusterfuzz._internal.bot.untrusted_runner import file_host
from clusterfuzz._internal.build_management import build_manager
from clusterfuzz._internal.crash_analysis.crash_result import CrashResult
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from clusterfuzz._internal.tests.test_libs import untrusted_runner_helpers
from clusterfuzz.fuzz import engine


class CreateTestcaseListFileTest(fake_filesystem_unittest.TestCase):
  """Tests for create_testcase_list_file."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)
    environment.set_value('FAIL_RETRIES', 1)

  def test_create(self):
    """Tests creation of testcase list file."""
    self.fs.create_file('/test/aa/bb.txt', contents='')
    self.fs.create_file('/test/aa/cc.txt', contents='')
    self.fs.create_file('/test/aa/aa/dd.txt', contents='')
    self.fs.create_file('/test/aa/aa/files.info', contents='')
    self.fs.create_file('/test/aa/bb/files.chrome.info', contents='')

    testcase_manager.create_testcase_list_file('/test/aa')

    testcase_list_file_path = '/test/aa/files.info'
    self.assertTrue(os.path.exists(testcase_list_file_path))

    expected_files_list = set([
        'bb.txt',
        'cc.txt',
        'aa/dd.txt',
    ])
    actual_files_list = set(open(testcase_list_file_path).read().splitlines())
    self.assertEqual(expected_files_list, actual_files_list)


# pylint: disable=protected-access
class UploadTestcaseOutputTest(fake_filesystem_unittest.TestCase):
  """Tests for logs uploading."""

  def setUp(self):
    """Setup for upload testcase output test."""
    test_helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)
    self.testcase_path = '/test/fuzzer/testcase'
    self.fs.create_file(self.testcase_path, contents='')

    environment.set_value('FUZZ_LOGS_BUCKET', 'fake-gcs-logs')
    environment.set_value('FUZZER_NAME', 'fuzzer')
    environment.set_value('JOB_NAME', 'job')
    environment.set_value('APP_REVISION', '123')

    # To be used for generation of date and time when uploading a log.
    fake_utcnow = datetime.datetime(2017, 5, 15, 16, 10, 28, 374119)

    # Original utcfromtimestamp needs to be preserved, as test_helpres does not
    # allow to patch only datetime.datetime.utcnow.
    orig_utcfromtimestamp = datetime.datetime.utcfromtimestamp

    test_helpers.patch(self, [
        'clusterfuzz._internal.build_management.revisions.get_component_range_list',
        'clusterfuzz._internal.google_cloud_utils.storage.write_data',
        'datetime.datetime',
    ])

    self.mock.datetime.utcnow.return_value = fake_utcnow
    self.mock.datetime.utcfromtimestamp.side_effect = orig_utcfromtimestamp
    self.mock.get_component_range_list.return_value = [{
        'component': 'Component',
        'link_text': 'REVISION',
    }, {
        'component': 'Component2',
        'link_text': 'REVISION2',
    }]

  def test_upload_with_timestamp_from_stats(self):
    """Log name should be generated using timestamp value from the stats."""
    mock_gsutil = mock.MagicMock()
    self.mock.write_data.return_value = mock_gsutil

    self.fs.create_file(
        self.testcase_path + '.stats2',
        contents='{"stat": 1000, "timestamp": 1472846341.017923, "kind": '
        '"TestcaseRun", "job": "job", "fuzzer": "fuzzer", '
        '"build_revision": 123}\n')

    environment.set_value('BOT_NAME', 'hostname.company.com')
    crash_result = CrashResult(
        return_code=1, crash_time=5, output='fake output')

    log = testcase_manager.prepare_log_for_upload(crash_result.get_stacktrace(),
                                                  crash_result.return_code)
    log_time = testcase_manager._get_testcase_time(self.testcase_path)
    testcase_manager.upload_log(log, log_time)

    # Date and time below is derived from 1472846341 timestamp value.
    self.mock.write_data.assert_called_once_with(
        b'Component revisions (build r123):\n'
        b'Component: REVISION\nComponent2: REVISION2\n\n'
        b'Bot name: hostname.company.com\n'
        b'Return code: 1\n\nfake output',
        'gs://fake-gcs-logs/fuzzer/job/2016-09-02/19:59:01:017923.log')

  def test_upload_with_hostname(self):
    """Log name should be generated using current (mocked) timestamp value."""
    mock_gsutil = mock.MagicMock()
    self.mock.write_data.return_value = mock_gsutil
    environment.set_value('BOT_NAME', 'hostname.company.com')

    self.fs.create_file(
        self.testcase_path + '.stats2',
        contents='{"stat": 1000, "kind": "TestcaseRun", "job": "job", '
        '"fuzzer": "fuzzer", "build_revision": 123}\n')

    crash_result = CrashResult(return_code=None, crash_time=None, output=None)
    log = testcase_manager.prepare_log_for_upload(crash_result.get_stacktrace(),
                                                  crash_result.return_code)
    log_time = testcase_manager._get_testcase_time(self.testcase_path)
    testcase_manager.upload_log(log, log_time)
    self.mock.write_data.assert_called_once_with(
        b'Component revisions (build r123):\n'
        b'Component: REVISION\nComponent2: REVISION2\n\n'
        b'Bot name: hostname.company.com\n'
        b'Return code: None\n\nNo output!',
        'gs://fake-gcs-logs/fuzzer/job/2017-05-15/16:10:28:374119.log')

  def test_upload_with_hostname_and_serial(self):
    """Log name should be generated using current (mocked) timestamp value."""
    mock_gsutil = mock.MagicMock()
    self.mock.write_data.return_value = mock_gsutil
    environment.set_value('BOT_NAME', 'hostname.company.com')
    environment.set_value('OS_OVERRIDE', 'ANDROID_KERNEL')
    environment.set_value('ANDROID_SERIAL', '123456789')

    self.fs.create_file(
        self.testcase_path + '.stats2',
        contents='{"stat": 1000, "kind": "TestcaseRun", "job": "job", '
        '"fuzzer": "fuzzer", "build_revision": 123}\n')

    crash_result = CrashResult(return_code=None, crash_time=None, output=None)
    log = testcase_manager.prepare_log_for_upload(crash_result.get_stacktrace(),
                                                  crash_result.return_code)
    log_time = testcase_manager._get_testcase_time(self.testcase_path)
    testcase_manager.upload_log(log, log_time)
    self.mock.write_data.assert_called_once_with(
        b'Component revisions (build r123):\n'
        b'Component: REVISION\nComponent2: REVISION2\n\n'
        b'Bot name: hostname.company.com\n'
        b'Device serial: 123456789\n'
        b'Return code: None\n\nNo output!',
        'gs://fake-gcs-logs/fuzzer/job/2017-05-15/16:10:28:374119.log')

  def test_upload_without_timestamp(self):
    """Log name should be generated using current (mocked) timestamp value."""
    mock_gsutil = mock.MagicMock()
    self.mock.write_data.return_value = mock_gsutil

    self.fs.create_file(
        self.testcase_path + '.stats2',
        contents='{"stat": 1000, "kind": "TestcaseRun", "job": "job", '
        '"fuzzer": "fuzzer", "build_revision": 123}\n')

    environment.set_value('BOT_NAME', 'hostname.company.com')
    crash_result = CrashResult(return_code=None, crash_time=None, output=None)
    log = testcase_manager.prepare_log_for_upload(crash_result.get_stacktrace(),
                                                  crash_result.return_code)
    log_time = testcase_manager._get_testcase_time(self.testcase_path)
    testcase_manager.upload_log(log, log_time)
    self.mock.write_data.assert_called_once_with(
        b'Component revisions (build r123):\n'
        b'Component: REVISION\nComponent2: REVISION2\n\n'
        b'Bot name: hostname.company.com\n'
        b'Return code: None\n\nNo output!',
        'gs://fake-gcs-logs/fuzzer/job/2017-05-15/16:10:28:374119.log')

  def test_upload_without_component_revisions(self):
    """Log should contain message on empty component revisions."""
    self.mock.get_component_range_list.return_value = []

    mock_gsutil = mock.MagicMock()
    self.mock.write_data.return_value = mock_gsutil

    self.fs.create_file(
        self.testcase_path + '.stats2',
        contents='{"stat": 1000, "timestamp": 1472846341.017923, "kind": '
        '"TestcaseRun", "job": "job", "fuzzer": "fuzzer", '
        '"build_revision": 123}\n')

    environment.set_value('BOT_NAME', 'hostname.company.com')
    crash_result = CrashResult(
        return_code=1, crash_time=5, output='fake output')
    log = testcase_manager.prepare_log_for_upload(crash_result.get_stacktrace(),
                                                  crash_result.return_code)
    log_time = testcase_manager._get_testcase_time(self.testcase_path)
    testcase_manager.upload_log(log, log_time)

    # Date and time below is derived from 1472846341 timestamp value.
    self.mock.write_data.assert_called_once_with(
        b'Component revisions (build r123):\n'
        b'Not available.\n\n'
        b'Bot name: hostname.company.com\n'
        b'Return code: 1\n\nfake output',
        'gs://fake-gcs-logs/fuzzer/job/2016-09-02/19:59:01:017923.log')


class ConvertDependencyUrlToLocalPathTest(unittest.TestCase):
  """Tests convert_dependency_url_to_local_path."""

  def setUp(self):
    test_helpers.patch_environ(self)
    environment.set_value('FUZZ_INPUTS', '/mnt/scratch0')

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.webserver.http_server.get_absolute_testcase_file',
        'clusterfuzz._internal.system.environment.platform',
        'clusterfuzz._internal.base.utils.normalize_path',
    ])
    self.mock.normalize_path.side_effect = lambda x: x

  def test_file_match_android(self):
    """Tests matching a file URL."""
    self.mock.platform.return_value = 'ANDROID'
    self.assertEqual(
        '/mnt/scratch0/test.html',
        testcase_manager.convert_dependency_url_to_local_path(
            'file:///sdcard/fuzzer-testcases/test.html'))
    self.mock.normalize_path.assert_called_once_with('/mnt/scratch0/test.html')

  def test_file_match_linux(self):
    """Tests matching a file URL."""
    self.mock.platform.return_value = 'LINUX'
    self.assertEqual(
        '/mnt/scratch0/test.html',
        testcase_manager.convert_dependency_url_to_local_path(
            'file:///mnt/scratch0/test.html'))
    self.mock.normalize_path.assert_called_once_with('/mnt/scratch0/test.html')

  def test_file_match_windows(self):
    """Tests matching a file URL."""
    self.mock.platform.return_value = 'WINDOWS'
    self.assertEqual(
        'C:/test/test.html',
        testcase_manager.convert_dependency_url_to_local_path(
            'file:///C:/test/test.html'))
    self.mock.normalize_path.assert_called_once_with('C:/test/test.html')

  def test_url_match_any_ip_and_port(self):
    """Tests matching a URL (any ip and port)."""
    self.mock.get_absolute_testcase_file.return_value = '/path'
    self.assertEqual(
        '/path',
        testcase_manager.convert_dependency_url_to_local_path(
            'http://10.240.1.237:8002/test/test.html'))
    self.mock.get_absolute_testcase_file.assert_called_once_with(
        '/test/test.html')
    self.mock.normalize_path.assert_called_once_with('/path')

  def test_url_match_localhost(self):
    """Tests matching a URL (localhost)."""
    self.mock.get_absolute_testcase_file.return_value = '/path'
    self.assertEqual(
        '/path',
        testcase_manager.convert_dependency_url_to_local_path(
            'http://localhost/test/test.html'))
    self.mock.get_absolute_testcase_file.assert_called_once_with(
        '/test/test.html')
    self.mock.normalize_path.assert_called_once_with('/path')

  def test_url_match_127_0_0_1(self):
    """Tests matching a URL (127.0.0.1)."""
    self.mock.get_absolute_testcase_file.return_value = '/path'
    self.assertEqual(
        '/path',
        testcase_manager.convert_dependency_url_to_local_path(
            'http://127.0.0.1/test/test.html'))
    self.mock.get_absolute_testcase_file.assert_called_once_with(
        '/test/test.html')
    self.mock.normalize_path.assert_called_once_with('/path')

  def test_url_match_127_0_0_1_with_port(self):
    """Tests matching a URL (127.0.0.1)."""
    self.mock.get_absolute_testcase_file.return_value = '/path'
    self.assertEqual(
        '/path',
        testcase_manager.convert_dependency_url_to_local_path(
            'http://127.0.0.1:8000/test/test.html'))
    self.mock.get_absolute_testcase_file.assert_called_once_with(
        '/test/test.html')
    self.mock.normalize_path.assert_called_once_with('/path')

  def test_not_match(self):
    """Tests not matching."""
    self.assertIsNone(
        testcase_manager.convert_dependency_url_to_local_path(
            'http://www.google.com/test.'))
    self.assertIsNone(
        testcase_manager.convert_dependency_url_to_local_path('random'))
    self.assertEqual(0, self.mock.normalize_path.call_count)


class GetResourcePathsTest(unittest.TestCase):
  """Tests get_resource_paths."""

  def setUp(self):
    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.testcase_manager.convert_dependency_url_to_local_path',
    ])
    self.mock.convert_dependency_url_to_local_path.side_effect = lambda x: x

  def test_get(self):
    """Tests getting resource paths."""
    # pylint: disable=line-too-long
    output = (
        '[54110:54110:0802/103321.557701:VERBOSE1:gles2_cmd_decoder.cc(3412)] GL_EXT_packed_depth_stencil supported.\n'
        '[54110:54110:0802/103321.577720:VERBOSE1:gles2_cmd_decoder.cc(3412)] GL_EXT_packed_depth_stencil supported.\n'
        '[54041:54075:0802/103321.618566:VERBOSE1:network_delegate.cc(31)] NetworkDelegate::NotifyBeforeURLRequest: '
        'file:///projects/chromium/src/third_party/WebKit/LayoutTests/paint/invalidation/resources/text-based-repaint.js\n'
        '[54041:54075:0802/103321.618677:VERBOSE1:network_delegate.cc(31)] NetworkDelegate::NotifyBeforeURLRequest: '
        'file:///projects/chromium/src/third_party/WebKit/LayoutTests/paint/invalidation/resources/text-based-repaint.js\n'
        '[54041:54075:0802/103321.618765:VERBOSE1:network_delegate.cc(31)] NetworkDelegate::NotifyBeforeURLRequest: '
        'file:///projects/chromium/src/third_party/WebKit/LayoutTests/resources/test-harness.js\n'
        '[1:1:0802/103321.658122:VERBOSE1:script_context.cc(110)] Created context:\r\n'
        '[3408:4060:0807/053137.661:54090359:VERBOSE1:network_delegate.cc(31)] NetworkDelegate::NotifyBeforeURLRequest: '
        'http://127.0.0.1:8000/fuzzer-common-data-bundles/webkit/layouttests/fast/canvas/fuzz-http-11.html\r\n'
        'V/chromium( 8530): [VERBOSE1:network_delegate.cc(31)] NetworkDelegate::NotifyBeforeURLRequest: https://en.m.wikipedia.org\n'
        '[36633:36655:1008/094348.879824:VERBOSE1:file_url_loader_factory.cc(441)] FileURLLoader::Start: file:///tmp/test.js'
    )

    result = testcase_manager.get_resource_paths(output)
    self.assertEqual(5, len(result))
    self.assertEqual(
        set([
            'file:///projects/chromium/src/third_party/WebKit/LayoutTests/paint/invalidation/resources/text-based-repaint.js',
            'file:///projects/chromium/src/third_party/WebKit/LayoutTests/resources/test-harness.js',
            'http://127.0.0.1:8000/fuzzer-common-data-bundles/webkit/layouttests/fast/canvas/fuzz-http-11.html',
            'https://en.m.wikipedia.org',
            'file:///tmp/test.js',
        ]), set(result))
    self.mock.convert_dependency_url_to_local_path.assert_has_calls([
        mock.call(
            'file:///projects/chromium/src/third_party/WebKit/LayoutTests/paint/invalidation/resources/text-based-repaint.js'
        ),
        mock.call(
            'file:///projects/chromium/src/third_party/WebKit/LayoutTests/resources/test-harness.js'
        ),
        mock.call(
            'http://127.0.0.1:8000/fuzzer-common-data-bundles/webkit/layouttests/fast/canvas/fuzz-http-11.html'
        ),
        mock.call('https://en.m.wikipedia.org')
    ])
    # pylint: enable=line-too-long


class GetCrashOutputTest(unittest.TestCase):
  """Tests _get_crash_output."""

  def test_none(self):
    self.assertEqual(None, testcase_manager._get_crash_output(None))  # pylint: disable=protected-access

  def test_no_end_marker(self):
    self.assertEqual('abc\ndef', testcase_manager._get_crash_output('abc\ndef'))  # pylint: disable=protected-access

  def test_end_marker(self):
    self.assertEqual(
        'abc\ndef\n',
        testcase_manager._get_crash_output(  # pylint: disable=protected-access
            'abc\ndef\nCRASH OUTPUT ENDS HERE\nghi'))


def mock_get_crash_data(output, symbolize_flag=True):  # pylint: disable=unused-argument
  """Mock get_crash_data."""
  if 'crash' in output:
    stack_analyzer_state = stacktraces.CrashInfo()
    stack_analyzer_state.crash_state = 'state'
    stack_analyzer_state.crash_type = 'Null-dereference'
    stack_analyzer_state.crash_stacktrace = output
    return stack_analyzer_state

  return stacktraces.CrashInfo()


@test_utils.with_cloud_emulators('datastore')
class TestcaseRunningTest(fake_filesystem_unittest.TestCase):
  """Tests for running testcases."""

  GREYBOX_FUZZER_NO_CRASH = ('Command: cmd\nTime ran: 0\n\noutput')
  GREYBOX_FUZZER_CRASH = 'Command: cmd\nTime ran: 1\n\ncrash'

  def setUp(self):
    """Setup for testcase running test."""
    test_helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)

    test_helpers.patch(self, [
        'clusterfuzz._internal.bot.fuzzers.engine_common.find_fuzzer_path',
        'clusterfuzz._internal.crash_analysis.stack_parsing.stack_analyzer.get_crash_data',
        'clusterfuzz._internal.system.process_handler.run_process',
        'clusterfuzz._internal.system.process_handler.'
        'terminate_stale_application_instances',
        'clusterfuzz.fuzz.engine.get',
        'clusterfuzz._internal.metrics.logs.log',
    ])

    os.environ['CRASH_RETRIES'] = '3'
    os.environ['FAIL_RETRIES'] = '3'
    os.environ['BOT_TMPDIR'] = '/bot/tmp'
    os.environ['TEST_TMPDIR'] = '/bot/tmp'
    os.environ['USER_PROFILE_ROOT_DIR'] = '/user-profiles'
    os.environ['APP_NAME'] = 'app_name'
    os.environ['APP_PATH'] = '/build_dir/app_name'
    os.environ['BUILD_DIR'] = os.environ['APP_DIR'] = '/build_dir'
    os.environ['CRASH_STACKTRACES_DIR'] = '/crashes'
    os.environ['INPUT_DIR'] = '/input'
    os.environ['FUZZER_DIR'] = '/fuzzer'
    os.environ['WARMUP_TIMEOUT'] = '120'
    os.environ['BOT_NAME'] = 'bot_name'

    data_types.FuzzTarget(engine='engine', project=None, binary='target').put()

    self.blackbox_testcase = data_types.Testcase(
        crash_state='state', overridden_fuzzer_name='fuzzer')
    self.greybox_testcase = data_types.Testcase(
        crash_state='state', overridden_fuzzer_name='engine_target')
    self.mock.find_fuzzer_path.return_value = '/build_dir/target'
    self.mock.run_process.return_value = (0, 0, 'output')
    self.mock.get.return_value = None

    self.mock.get_crash_data.side_effect = mock_get_crash_data
    self.fs.create_file('/flags-testcase', contents='-arg1 -arg2')
    self.fs.create_dir('/bot/tmp')

  def test_test_for_crash_with_retries_blackbox_fail(self):
    """Test test_for_crash_with_retries failing to reproduce a crash
    (blackbox)."""
    crash_result = testcase_manager.test_for_crash_with_retries(
        self.blackbox_testcase, '/fuzz-testcase', 10)
    self.assertEqual(0, crash_result.return_code)
    self.assertEqual(0, crash_result.crash_time)
    self.assertEqual('output', crash_result.output)
    self.assertEqual(3, self.mock.run_process.call_count)
    self.mock.run_process.assert_has_calls([
        mock.call(
            '/build_dir/app_name -arg1 -arg2',
            current_working_directory='/build_dir',
            gestures=[],
            timeout=120),
        mock.call(
            '/build_dir/app_name -arg1 -arg2',
            current_working_directory='/build_dir',
            gestures=[],
            timeout=10),
        mock.call(
            '/build_dir/app_name -arg1 -arg2',
            current_working_directory='/build_dir',
            gestures=[],
            timeout=10),
    ])
    self.mock.log.assert_has_calls([
        mock.call('No crash occurred (round 1).', output='output'),
        mock.call('No crash occurred (round 2).', output='output'),
        mock.call('No crash occurred (round 3).', output='output'),
        mock.call("Didn't crash at all.")
    ])

  def test_test_for_crash_with_retries_greybox_fail(self):
    """Test test_for_crash_with_retries failing to reproduce a crash
    (greybox)."""
    mock_engine = mock.Mock()
    mock_engine.reproduce.return_value = engine.ReproduceResult(['cmd'], 0, 0,
                                                                'output')
    self.mock.get.return_value = mock_engine

    crash_result = testcase_manager.test_for_crash_with_retries(
        self.greybox_testcase, '/fuzz-testcase', 10)
    self.assertEqual(0, crash_result.return_code)
    self.assertEqual(0, crash_result.crash_time)
    self.assertEqual(self.GREYBOX_FUZZER_NO_CRASH, crash_result.output)
    self.assertEqual(3, mock_engine.reproduce.call_count)
    mock_engine.reproduce.assert_has_calls([
        mock.call('/build_dir/target', '/fuzz-testcase', ['-arg1', '-arg2'],
                  120),
        mock.call('/build_dir/target', '/fuzz-testcase', ['-arg1', '-arg2'],
                  10),
        mock.call('/build_dir/target', '/fuzz-testcase', ['-arg1', '-arg2'],
                  10),
    ])
    self.mock.log.assert_has_calls(
        [
            mock.call(
                'No crash occurred (round 1).',
                output=self.GREYBOX_FUZZER_NO_CRASH),
            mock.call(
                'No crash occurred (round 2).',
                output=self.GREYBOX_FUZZER_NO_CRASH),
            mock.call(
                'No crash occurred (round 3).',
                output=self.GREYBOX_FUZZER_NO_CRASH),
            mock.call("Didn't crash at all.")
        ])

  def test_test_for_crash_with_retries_blackbox_succeed(self):
    """Test test_for_crash_with_retries reproducing a crash (blackbox)."""
    self.mock.run_process.side_effect = [
        (0, 0, 'output'),
        (1, 1, 'crash'),
    ]

    crash_result = testcase_manager.test_for_crash_with_retries(
        self.blackbox_testcase, '/fuzz-testcase', 10)
    self.assertEqual(1, crash_result.return_code)
    self.assertEqual(1, crash_result.crash_time)
    self.assertEqual('crash', crash_result.output)
    self.assertEqual(2, self.mock.run_process.call_count)

    self.mock.run_process.assert_has_calls([
        mock.call(
            '/build_dir/app_name -arg1 -arg2',
            current_working_directory='/build_dir',
            gestures=[],
            timeout=120),
        mock.call(
            '/build_dir/app_name -arg1 -arg2',
            current_working_directory='/build_dir',
            gestures=[],
            timeout=10),
    ])
    self.mock.log.assert_has_calls([
        mock.call('No crash occurred (round 1).', output='output'),
        mock.call(
            'Crash occurred in 1 seconds (round 2). State:\nstate',
            output='crash'),
        mock.call('Crash stacktrace is similar to original stacktrace.')
    ])

  def test_test_for_crash_with_retries_blackbox_succeed_no_comparison(self):
    """Test test_for_crash_with_retries reproducing a crash with compare_crash
    set to False (blackbox)."""
    self.mock.run_process.side_effect = [
        (0, 0, 'output'),
        (1, 1, 'crash'),
    ]

    crash_result = testcase_manager.test_for_crash_with_retries(
        self.blackbox_testcase, '/fuzz-testcase', 10, compare_crash=False)
    self.assertEqual(1, crash_result.return_code)
    self.assertEqual(1, crash_result.crash_time)
    self.assertEqual('crash', crash_result.output)
    self.assertEqual(2, self.mock.run_process.call_count)

    self.mock.run_process.assert_has_calls([
        mock.call(
            '/build_dir/app_name -arg1 -arg2',
            current_working_directory='/build_dir',
            gestures=[],
            timeout=120),
        mock.call(
            '/build_dir/app_name -arg1 -arg2',
            current_working_directory='/build_dir',
            gestures=[],
            timeout=10),
    ])
    self.mock.log.assert_has_calls([
        mock.call('No crash occurred (round 1).', output='output'),
        mock.call(
            'Crash occurred in 1 seconds (round 2). State:\nstate',
            output='crash'),
        mock.call('Crash stacktrace comparison skipped.')
    ])

  def test_test_for_crash_with_retries_greybox_succeed(self):
    """Test test_for_crash_with_retries reproducing a crash (greybox)."""
    mock_engine = mock.Mock()
    mock_engine.reproduce.side_effect = [
        engine.ReproduceResult(['cmd'], 0, 0, 'output'),
        engine.ReproduceResult(['cmd'], 1, 1, 'crash'),
    ]
    self.mock.get.return_value = mock_engine

    crash_result = testcase_manager.test_for_crash_with_retries(
        self.greybox_testcase, '/fuzz-testcase', 10)
    self.assertEqual(1, crash_result.return_code)
    self.assertEqual(1, crash_result.crash_time)
    self.assertEqual(self.GREYBOX_FUZZER_CRASH, crash_result.output)
    self.assertEqual(2, mock_engine.reproduce.call_count)
    mock_engine.reproduce.assert_has_calls([
        mock.call('/build_dir/target', '/fuzz-testcase', ['-arg1', '-arg2'],
                  120),
        mock.call('/build_dir/target', '/fuzz-testcase', ['-arg1', '-arg2'],
                  10),
    ])
    self.mock.log.assert_has_calls([
        mock.call(
            'No crash occurred (round 1).',
            output=self.GREYBOX_FUZZER_NO_CRASH),
        mock.call(
            'Crash occurred in 1 seconds (round 2). State:\nstate',
            output=self.GREYBOX_FUZZER_CRASH),
        mock.call('Crash stacktrace is similar to original stacktrace.')
    ])

  def test_test_for_crash_with_retries_greybox_succeed_no_comparison(self):
    """Test test_for_crash_with_retries reproducing a crash with compare_crash
    set to False (greybox)."""
    mock_engine = mock.Mock()
    mock_engine.reproduce.side_effect = [
        engine.ReproduceResult(['cmd'], 0, 0, 'output'),
        engine.ReproduceResult(['cmd'], 1, 1, 'crash'),
    ]
    self.mock.get.return_value = mock_engine

    crash_result = testcase_manager.test_for_crash_with_retries(
        self.greybox_testcase, '/fuzz-testcase', 10, compare_crash=False)
    self.assertEqual(1, crash_result.return_code)
    self.assertEqual(1, crash_result.crash_time)
    self.assertEqual(self.GREYBOX_FUZZER_CRASH, crash_result.output)
    self.assertEqual(2, mock_engine.reproduce.call_count)
    mock_engine.reproduce.assert_has_calls([
        mock.call('/build_dir/target', '/fuzz-testcase', ['-arg1', '-arg2'],
                  120),
        mock.call('/build_dir/target', '/fuzz-testcase', ['-arg1', '-arg2'],
                  10),
    ])
    self.mock.log.assert_has_calls([
        mock.call(
            'No crash occurred (round 1).',
            output=self.GREYBOX_FUZZER_NO_CRASH),
        mock.call(
            'Crash occurred in 1 seconds (round 2). State:\nstate',
            output=self.GREYBOX_FUZZER_CRASH),
        mock.call('Crash stacktrace comparison skipped.')
    ])

  def test_test_for_crash_with_retries_greybox_legacy(self):
    """Test test_for_crash_with_retries reproducing a legacy crash (greybox)."""
    mock_engine = mock.Mock()
    mock_engine.reproduce.side_effect = [
        engine.ReproduceResult(['cmd'], 1, 1, 'crash'),
    ]
    self.mock.get.return_value = mock_engine

    with open('/flags-testcase', 'w') as f:
      f.write('%TESTCASE% target -arg1 -arg2')

    testcase_manager.test_for_crash_with_retries(self.greybox_testcase,
                                                 '/fuzz-testcase', 10)
    mock_engine.reproduce.assert_has_calls([
        mock.call('/build_dir/target', '/fuzz-testcase', ['-arg1', '-arg2'],
                  120),
    ])
    self.mock.log.assert_has_calls([
        mock.call(
            'Crash occurred in 1 seconds (round 1). State:\nstate',
            output=self.GREYBOX_FUZZER_CRASH),
        mock.call('Crash stacktrace is similar to original stacktrace.')
    ])

  def test_test_for_reproducibility_blackbox_succeed(self):
    """Test test_for_reproducibility with success on all runs (blackbox)."""
    self.mock.run_process.return_value = (1, 1, 'crash')
    result = testcase_manager.test_for_reproducibility(
        'fuzzer',
        'fuzzer',
        '/fuzz-testcase',
        'state',
        expected_security_flag=False,
        test_timeout=10,
        http_flag=False,
        gestures=None)
    self.assertTrue(result)

    # Only 2/3 runs needed to verify reproducibility.
    self.assertEqual(2, self.mock.run_process.call_count)
    self.mock.log.assert_has_calls([
        mock.call(
            'Crash occurred in 1 seconds (round 1). State:\nstate',
            output='crash'),
        mock.call(
            'Crash occurred in 1 seconds (round 2). State:\nstate',
            output='crash'),
        mock.call('Crash is reproducible.'),
    ])

  def test_test_for_reproducibility_blackbox_succeed_after_multiple_tries(self):
    """Test test_for_reproducibility with failure on first run and then succeed
    on remaining runs (blackbox)."""
    self.mock.run_process.side_effect = [
        (0, 0, 'output'),
        (1, 1, 'crash'),
        (1, 1, 'crash'),
    ]
    result = testcase_manager.test_for_reproducibility(
        'fuzzer',
        'fuzzer',
        '/fuzz-testcase',
        'state',
        expected_security_flag=False,
        test_timeout=10,
        http_flag=False,
        gestures=None)
    self.assertTrue(result)

    self.assertEqual(3, self.mock.run_process.call_count)
    self.mock.log.assert_has_calls([
        mock.call('No crash occurred (round 1).', output='output'),
        mock.call(
            'Crash occurred in 1 seconds (round 2). State:\nstate',
            output='crash'),
        mock.call(
            'Crash occurred in 1 seconds (round 3). State:\nstate',
            output='crash'),
        mock.call('Crash is reproducible.'),
    ])

  def test_test_for_reproducibility_greybox_succeed(self):
    """Test test_for_reproducibility with success on all runs (greybox)."""
    mock_engine = mock.Mock()
    mock_engine.reproduce.return_value = engine.ReproduceResult(['cmd'], 1, 1,
                                                                'crash')
    self.mock.get.return_value = mock_engine

    result = testcase_manager.test_for_reproducibility(
        'engine',
        'engine_target',
        '/fuzz-testcase',
        'state',
        expected_security_flag=False,
        test_timeout=10,
        http_flag=False,
        gestures=None)
    self.assertTrue(result)

    # Only 2/3 runs needed to verify reproducibility.
    self.assertEqual(2, mock_engine.reproduce.call_count)
    self.mock.log.assert_has_calls([
        mock.call(
            'Crash occurred in 1 seconds (round 1). State:\nstate',
            output=self.GREYBOX_FUZZER_CRASH),
        mock.call(
            'Crash occurred in 1 seconds (round 2). State:\nstate',
            output=self.GREYBOX_FUZZER_CRASH),
        mock.call('Crash is reproducible.'),
    ])

  def test_test_for_reproducibility_greybox_succeed_after_multiple_tries(self):
    """Test test_for_reproducibility with with failure on first run and then
    succeed on remaining runs  (greybox)."""
    mock_engine = mock.Mock()
    mock_engine.reproduce.side_effect = [
        engine.ReproduceResult(['cmd'], 0, 0, 'output'),
        engine.ReproduceResult(['cmd'], 1, 1, 'crash'),
        engine.ReproduceResult(['cmd'], 1, 1, 'crash'),
    ]
    self.mock.get.return_value = mock_engine

    result = testcase_manager.test_for_reproducibility(
        'engine',
        'engine_target',
        '/fuzz-testcase',
        'state',
        expected_security_flag=False,
        test_timeout=10,
        http_flag=False,
        gestures=None)
    self.assertTrue(result)

    self.assertEqual(3, mock_engine.reproduce.call_count)
    self.mock.log.assert_has_calls([
        mock.call(
            'No crash occurred (round 1).',
            output=self.GREYBOX_FUZZER_NO_CRASH),
        mock.call(
            'Crash occurred in 1 seconds (round 2). State:\nstate',
            output=self.GREYBOX_FUZZER_CRASH),
        mock.call(
            'Crash occurred in 1 seconds (round 3). State:\nstate',
            output=self.GREYBOX_FUZZER_CRASH),
        mock.call('Crash is reproducible.'),
    ])


class UntrustedEngineReproduceTest(
    untrusted_runner_helpers.UntrustedRunnerIntegrationTest):
  """Engine reproduction tests for untrusted."""

  def setUp(self):
    """Set up."""
    super().setUp()
    environment.set_value('JOB_NAME', 'libfuzzer_asan_job')

    job = data_types.Job(
        name='libfuzzer_asan_job',
        environment_string=(
            'RELEASE_BUILD_BUCKET_PATH = '
            'gs://clusterfuzz-test-data/test_libfuzzer_builds/'
            'test-libfuzzer-build-([0-9]+).zip\n'
            'REVISION_VARS_URL = https://commondatastorage.googleapis.com/'
            'clusterfuzz-test-data/test_libfuzzer_builds/'
            'test-libfuzzer-build-%s.srcmap.json\n'))
    job.put()

    self.temp_dir = tempfile.mkdtemp(dir=environment.get_value('FUZZ_INPUTS'))

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.temp_dir, ignore_errors=True)

  def test_reproduce(self):
    """Test reproduce."""
    testcase_file_path = os.path.join(self.temp_dir, 'testcase')
    with open(testcase_file_path, 'wb') as f:
      f.write(b'EEE')

    self._setup_env(job_type='libfuzzer_asan_job')

    build_manager.setup_build()
    result = testcase_manager.engine_reproduce(
        libfuzzer_engine.LibFuzzerEngine(), 'test_fuzzer', testcase_file_path,
        [], 30)

    self.assertEqual([
        os.path.join(environment.get_value('BUILD_DIR'), 'test_fuzzer'),
        '-runs=100',
        file_host.rebase_to_worker_root(testcase_file_path)
    ], result.command)
    self.assertEqual(result.return_code,
                     libfuzzer_constants.TARGET_ERROR_EXITCODE)
    self.assertGreater(result.time_executed, 0)
    self.assertIn('Running 1 inputs 100 time(s) each', result.output)
    self.assertIn('AddressSanitizer: SEGV on unknown address 0x000000000000',
                  result.output)

  def test_target_not_found(self):
    """Test target not found."""
    testcase_file_path = os.path.join(self.temp_dir, 'testcase')
    with open(testcase_file_path, 'wb') as f:
      f.write(b'EEE')

    self._setup_env(job_type='libfuzzer_asan_job')

    build_manager.setup_build()
    with self.assertRaises(testcase_manager.TargetNotFoundError):
      testcase_manager.engine_reproduce(libfuzzer_engine.LibFuzzerEngine(),
                                        'does_not_exist', testcase_file_path,
                                        [], 30)


class GetCommandLineFlagsTest(fake_filesystem_unittest.TestCase):
  """get_command_line_flags tests."""

  def setUp(self):
    test_helpers.patch_environ(self)
    test_utils.set_up_pyfakefs(self)

    os.environ['FAIL_WAIT'] = '0'
    os.environ['FAIL_RETRIES'] = '1'

    self.fs.create_file('/fuzz-testcase')

  def test_both_args_and_additional(self):
    """Test both APP_ARGS and additional args."""
    os.environ['APP_ARGS'] = 'arg1'
    self.fs.create_file('/flags-testcase', contents='arg2')
    self.assertEqual('arg1 arg2',
                     testcase_manager.get_command_line_flags('/fuzz-testcase'))

  def test_only_args(self):
    """Test both APP_ARGS and additional args."""
    os.environ['APP_ARGS'] = 'arg1'
    self.assertEqual('arg1',
                     testcase_manager.get_command_line_flags('/fuzz-testcase'))

  def test_only_additional(self):
    """Test both APP_ARGS and additional args."""
    self.fs.create_file('/flags-testcase', contents='arg2')
    self.assertEqual('arg2',
                     testcase_manager.get_command_line_flags('/fuzz-testcase'))

  def test_no_args_and_additional(self):
    """Test both APP_ARGS and additional args."""
    self.assertEqual('',
                     testcase_manager.get_command_line_flags('/fuzz-testcase'))
